from pydantic import BaseModel

# RCIC
class Rcic(BaseModel):
    first_name:str
    last_name:str
    company:str
    sex:str
    rcic_number:str
    
    @property
    def name(self):
        return self.first_name+' '+self.last_name
    
    


from model.tr.pdfformmodel import PdfFormModel
from model.tr.d5645 import data_model, convert_model,remove_model

class F5645Model(PdfFormModel):
    def __init__(self,xml_file):
        # pass the globals variables to base class
        model={'data_model':data_model,'convert_model':convert_model,'remove_model':remove_model}
        super().__init__(xml_file,model)
        self._reorder()
        
    def _getSpecial(self):
        # child 
        children=self.tree.iterfind('page1/SectionB/Child')
        children_list=[]
        for child in children:
            child_item=self._getSubItem(child,data_model['children'])
            children_list.append(child_item)
        self.data['children']=children_list
        
        #siblings
        siblings=self.tree.iterfind('page1/SectionC/Child')
        siblings_list=[]
        for sibling in siblings:
            sibling_item=self._getSubItem(sibling,data_model['siblings'])
            siblings_list.append(sibling_item)
        self.data['siblings']=siblings_list
        

    
    def _assembl(self):
        for sheet,value in self.data.items():
            if sheet=='applicant': continue
            if isinstance(value,dict):
                self.data[sheet]['accompany_to_canada']=value['accompany_to_canada_yes']
            elif isinstance(value,list):
                for index,d in enumerate(value):
                    self.data[sheet][index]['accompany_to_canada']=value[index]['accompany_to_canada_yes']
            else:
                pass
        
        self.data['family']=[
            {
                'relationship':'applicant',
                **self.data['applicant']
            },
            {
                'relationship':'spouse',
                **self.data['spouse']
            },
            {
                'relationship':'mother',
                **self.data['mother']
            },
            {
                'relationship':'father',
                **self.data['father']
            },
            *self.data['children'],
            *self.data['siblings']
        ]
        # re-order the dict to match execl
    
    def _remove(self):
        popup_pairs={
            'spouse':['accompany_to_canada_yes','accompany_to_canada_no'],
            'mother':['accompany_to_canada_yes','accompany_to_canada_no'],
            'father':['accompany_to_canada_yes','accompany_to_canada_no']
        }
        for k,v in popup_pairs.items():
            [self.data[k].pop(item) for item in v]
        [self.data.pop(key)  for key in ['children','siblings','spouse','applicant','mother','father']  ]
        for index,f in enumerate(self.data['family']):
            if f.get('relationship')!='applicant':
                self.data['family'][index].pop('accompany_to_canada_yes',None)
                self.data['family'][index].pop('accompany_to_canada_no',None)
        
    def _reorder(self):
        #get name not none members
        members=[ m for m in self.data['family'] if m.get('name')!=None]
        # get first last name and native language first last name
        for member in members:
            #get all names
            names=member['name'].replace(',',' ') # remove ',' 
            names=' '.join(names.split()) # replace multiple space with one. 
            names=names.split(' ') # split names 
            native_last_name,native_first_name=names[-1][0],names[-1][1:] # get native language name
            last_name,first_name=names[:-1][0],''.join(names[:-1][1:]) #get first and last name
            member['last_name']=last_name
            member['first_name']=first_name
            member['native_last_name']=native_last_name
            member['native_first_name']=native_first_name
            member.pop('name')
        # or-order the dict
        temp_list=[]
        for member in members:
            temp={}
            for k in ['last_name','first_name','native_last_name','native_first_name','marital_status','date_of_birth','birth_country','address','occupation','relationship','accompany_to_canada']:
                temp[k]=member.get(k)
            temp_list.append(temp)
        self.data['family']=temp_list
            



from typing import List
from model.common.commonmodel import CommonModel
from model.recruit.jobaddata import JobOffer,General,ErAddress,ErAddresses
from model.common.wordmaker import WordMaker
import os

class JobadModel(CommonModel):
    general:General
    joboffer:JobOffer
    eraddress:List[ErAddress]
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/er.xlsx'
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())
        
class JobadModelDocxAdapater():
    """This is an adapater to bridging job ad model data and docx data
    """

    def __init__(self,jobad_obj: JobadModel):
        # get original obj, which will be used to generate some value based on it's object methods. 
        # 此处用来处理list里面的一些内容。 
        self.jobad_obj=jobad_obj
        addresses=ErAddresses(self.jobad_obj.eraddress)
        self.jobad_obj.eraddress=addresses.working

    def make(self,output_docx):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/word/jobad.docx"            
        wm=WordMaker(template_path,self.jobad_obj,output_docx)
        wm.make()




from pydantic import BaseModel,validator
from datetime import date,timedelta
from typing import Optional,List
from model.common.mixins import DurationMixin
from model.common.utils import makeList,Duration
from dateutil.parser import parse,ParserError

class EmploymentBase(BaseModel,DurationMixin):
    start_date:date
    end_date:Optional[date]
    job_title:str
    noc_code:str
    weekly_hours:float
    company:str
    city:str
    province:str
    country:str

    @validator("end_date")
    def endDateBigger(cls,end_date,values):
        start_date=values.get('start_date')
        the_date=end_date if end_date else date.today()
        if (the_date-start_date).days<=0:
            raise ValueError(f'End date {the_date} is earlier than start date {start_date}')
        end_date=end_date if end_date else "Present"
        return end_date
    
    @property
    def part_time(self):
        return "Part time" if self.weekly_hours<30 else ""
    
    @property
    def is_current(self):
        return True if self.end_date=='Present' else False

    @property
    def years(self):
        return Duration(self.start_date,self.end_date).years
    
    @property
    def months(self):
        return Duration(self.start_date,self.end_date).years
    
    def yearsOnDate(self,end_date):
        return Duration(self.start_date,self.end_date).yearsOnDate(end_date)
    
    def monthsOnDate(self,end_date):
        return Duration(self.start_date,self.end_date).monthsOnDate(end_date)
    
    @property
    def start_date_mmyyyy(self):
        return self.start_date.strftime('%b %Y')

    @property
    def end_date_mmyyyy(self):
        return self.end_date.strftime('%b %Y') if self.end_date else "Present"




country = {
        "Afghanistan": "252",
        "Aland Island": "401",
        "Albania": "081",
        "Algeria": "131",
        "Andorra": "082",
        "Angola": "151",
        "Anguilla": "620",
        "Antigua And Barbuda": "621",
        "Argentina": "703",
        "Armenia": "049",
        "Aruba": "658",
        "Australia": "305",
        "Austria": "011",
        "Azerbaijan": "050",
        "Bahamas": "622",
        "Bahrain": "253",
        "Bailwick of Jersey": "412",
        "Bangladesh": "212",
        "Barbados": "610",
        "Belarus": "051",
        "Belgium": "012",
        "Belize": "541",
        "Benin": "160",
        "Bermuda": "601",
        "Bhutan": "254",
        "Bolivia": "751",
        "Bonaire, Sint Eustatius, Saba": "402",
        "Bosnia and Herzegovina": "048",
        "Botswana": "153",
        "Bouvet Island": "403",
        "Brazil": "709",
        "British Indian Ocean Territory": "404",
        "British Virgin Islands": "633",
        "Brunei Darussalam": "255",
        "Bulgaria": "083",
        "Burkina Faso": "188",
        "Burma (Myanmar)": "241",
        "Burundi": "154",
        "Cabo Verde": "911",
        "Cambodia": "256",
        "Cameroon": "155",
        "Canada": "511",
        "Canary Islands": "039",
        "Cayman Islands": "624",
        "Central African Republic": "157",
        "Chad": "156",
        "Channel Islands": "009",
        "Chile": "721",
        "China": "202",
        "China (Hong Kong SAR)": "200",
        "China (Macao SAR)": "198",
        "Christmas Island": "405",
        "Colombia": "722",
        "Comoros": "905",
        "Cook Islands": "840",
        "Costa Rica": "542",
        "Croatia": "043",
        "Cuba": "650",
        "Cura\u00e7ao": "406",
        "Cyprus": "221",
        "Czech Republic": "015",
        "Czechoslovakia": "014",
        "Democratic Rep. of Congo": "158",
        "Denmark": "017",
        "Djibouti": "183",
        "Dominica": "625",
        "Dominican Republic": "651",
        "East Timor": "916",
        "Ecuador": "753",
        "Egypt": "101",
        "El Salvador": "543",
        "England": "002",
        "Equatorial Guinea": "178",
        "Eritrea": "162",
        "Estonia": "018",
        "Ethiopia": "161",
        "Falkland Islands": "912",
        "Faroe Islands": "408",
        "Federated States of Micronesia": "835",
        "Fiji": "801",
        "Finland": "021",
        "Fr. South. and Antarctic Lands": "821",
        "France": "022",
        "French Guiana": "754",
        "French Polynesia": "845",
        "Gabon": "163",
        "Gambia": "164",
        "Georgia": "052",
        "Germany, Federal Republic Of": "024",
        "Ghana": "165",
        "Gibraltar": "084",
        "Greece": "025",
        "Greenland": "521",
        "Grenada": "626",
        "Guadeloupe": "653",
        "Guam": "832",
        "Guatemala": "544",
        "Guernsey": "409",
        "Guinea": "166",
        "Guinea-Bissau": "167",
        "Guyana": "711",
        "Haiti": "654",
        "Heard and MacDonald Islands": "410",
        "Honduras": "545",
        "Hong Kong": "204",
        "Hungary": "026",
        "Iceland": "085",
        "India": "205",
        "Indonesia": "222",
        "Iran": "223",
        "Iraq": "224",
        "Ireland": "027",
        "Isle of Man": "411",
        "Israel": "206",
        "Italy": "028",
        "Ivory Coast": "169",
        "Jamaica": "602",
        "Japan": "207",
        "Jordan": "225",
        "Kampuchea Democratic Rep.": "211",
        "Kazakhstan": "053",
        "Kenya": "132",
        "Kiribati": "831",
        "Korea, North (DPRK)": "257",
        "Korea, South": "258",
        "Kosovo": "064",
        "Kuwait": "226",
        "Kyrgyzstan": "054",
        "Laos": "260",
        "Latvia": "019",
        "Lebanon": "208",
        "Lesotho": "152",
        "Liberia": "170",
        "Libya": "171",
        "Liechtenstein": "086",
        "Lithuania": "020",
        "Luxembourg": "013",
        "Macao": "261",
        "Macedonia": "070",
        "Madagascar": "172",
        "Madeira": "036",
        "Malawi": "111",
        "Malaysia": "242",
        "Maldives": "901",
        "Mali": "173",
        "Malta": "030",
        "Marshall Islands": "834",
        "Martinique": "655",
        "Mauritania": "174",
        "Mauritius": "902",
        "Mayotte": "906",
        "Mexico": "501",
        "Moldova": "055",
        "Monaco": "087",
        "Mongolia": "262",
        "Montenegro": "063",
        "Montserrat": "627",
        "Morocco": "133",
        "Mozambique": "175",
        "Namibia": "122",
        "Nauru": "341",
        "Nepal": "264",
        "Netherlands Antilles, The": "652",
        "Netherlands, The": "031",
        "Nevis": "628",
        "New Caledonia": "822",
        "New Zealand": "339",
        "Nicaragua": "546",
        "Niger": "176",
        "Nigeria": "177",
        "Niue": "414",
        "Northern Ireland": "006",
        "Northern Mariana Islands": "830",
        "Norway": "032",
        "Oman": "263",
        "Pakistan": "209",
        "Palestinian Authority": "213",
        "Panama": "547",
        "Papua New Guinea": "342",
        "Paraguay": "755",
        "Peru": "723",
        "Philippines": "227",
        "Pitcairn Islands": "842",
        "Poland": "033",
        "Portugal": "034",
        "Puerto Rico": "656",
        "Qatar": "265",
        "Republic of Congo": "159",
        "Republic of Palau": "836",
        "R\u00e9union": "903",
        "Romania": "088",
        "Russia": "056",
        "Rwanda": "179",
        "Saint Helena": "915",
        "Saint Kitts and Nevis": "629",
        "Saint Lucia": "630",
        "Saint Pierre and Miquelon": "531",
        "Saint-Barthelemy": "407",
        "Saint-Martin": "415",
        "Samoa": "844",
        "Samoa, American": "843",
        "San Marino": "089",
        "Sao Tome and Principe": "914",
        "Saudi Arabia": "231",
        "Scotland": "007",
        "Senegal": "180",
        "Serbia And Montenegro": "061",
        "Serbia, Republic Of": "062",
        "Seychelles": "904",
        "Sierra Leone": "181",
        "Singapore": "246",
        "Sint-Maarten": "416",
        "Slovakia": "016",
        "Slovenia": "047",
        "Solomon Islands": "824",
        "Somalia": "182",
        "South Africa, Republic Of": "121",
        "South Sudan": "189",
        "Spain": "037",
        "Sri Lanka": "201",
        "St. Vincent and the Grenadines": "631",
        "Sudan": "185",
        "Suriname": "752",
        "Swaziland": "186",
        "Sweden": "040",
        "Switzerland": "041",
        "Syria": "210",
        "Taiwan": "203",
        "Tajikistan": "057",
        "Tanzania": "130",
        "Thailand": "267",
        "Tibet (Autonomous Region)": "268",
        "Togo": "187",
        "Tokelau": "417",
        "Tonga": "846",
        "Trinidad and Tobago": "605",
        "Tunisia": "135",
        "Turkey": "045",
        "Turkmenistan": "058",
        "Turks and Caicos Islands": "632",
        "Tuvalu": "826",
        "U.S. Minor outlying Islands": "418",
        "Uganda": "136",
        "Ukraine": "059",
        "Union Of Soviet Socialist Rep": "042",
        "United Arab Emirates": "280",
        "United States of America": "461",
        "Unknown": "000",
        "Uruguay": "724",
        "Uzbekistan": "060",
        "Vanuatu": "823",
        "Vatican City State": "090",
        "Venezuela": "725",
        "Vietnam": "270",
        "Virgin Islands, U.S.": "657",
        "Wales": "008",
        "Wallis and Futuna Is., Terr.": "841",
        "Western Sahara": "184",
        "Yemen": "273",
        "Yugoslavia": "044",
        "Zambia": "112",
        "Zimbabwe": "113"
    } # total 262



from pydantic import BaseModel,EmailStr
from typing import Optional,List
from collections import namedtuple
from model.common.mixins import EqMixin

class ContactBase(BaseModel):
    contact_variable:Optional[str]
    contact_type:Optional[str]
    last_name:Optional[str]
    first_name:Optional[str]
    phone:Optional[str]
    email:Optional[EmailStr]

    class Config:
        anystr_lower=True 

class Contacts(object):
    def __init__(self, contacts: List[ContactBase]):
        self.contacts=contacts
    
    def _specific_contact(self,v_type):
        contact=[contact for contact in self.contacts if contact.contact_variable==v_type]
        return contact[0] if contact else None
    
    @property
    def primary(self):
        return self._specific_contact('primary')
    
    @property
    def second(self):
        return self._specific_contact('second')
    
    @property
    def preferredContact(self):
        for contact_type in ['primary','second']:
            contact=self._specific_contact(contact_type)
            if contact: return contact
    




from termcolor import colored
from functools import reduce
from typing import List,Optional
from model.common.address import Address
from model.common.phone import Phone
from model.common.trperson import COR, PersonId,Personal,Marriage,Education,Employment,Travel,Family
from model.common.tr import TrCaseIn,TrBackground, VrInCanada
from model.common.commonmodel import CommonModel


class M5708Model(CommonModel):
    personal:Personal
    marriage:Marriage
    personid:List[PersonId]
    address:List[Address]
    education:List[Education]
    employment:List[Employment]
    travel:List[Travel]
    family:List[Family]
    phone:List[Phone]
    cor:List[COR]
    trcasein:TrCaseIn
    vrincanada:VrInCanada
    trbackground:TrBackground
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            import os
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/tr.xlsx',
                path+'/template/excel/pa.xlsx'
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        super().__init__(excels,output_excel_file,globals())




"""
This is a tool set for pydantic validation shared by all models
"""
from datetime import date
from dateutil.parser import parse,ParserError

def normalize(name: str):
    if name:
        return ' '.join((word.capitalize()) for word in name.split(' '))
    return ''

def makeList(name:str):
    if name:
        return name.split('\n')
    return []

def trimString(text:str):
    return "" if text==None else text.strip()

class Duration():
    def __init__(self,start_date,end_date):
        try:
            self.start_date=start_date if isinstance(start_date,date) else parse(start_date)
            self.end_date=end_date if isinstance(end_date,date) else parse(end_date)
        except ParserError as err:
            raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        
    @property
    def years(self):
        return self.end_date.year - self.start_date.year - ((self.end_date.month, self.end_date.day) < (self.start_date.month, self.start_date.day))
    
    @property
    def months(self):
        years2months=(self.end_date.year - self.start_date.year)*12
        months2months=self.end_date.month-self.start_date.month
        return years2months+months2months
    
    # get years on a specific date
    def yearsOnDate(self,end_date= date.today()):
        if not isinstance(end_date,date):
            try:
                end_date=parse(end_date)
                return end_date.year - self.start_date.year - ((end_date.month, end_date.day) < (self.start_date.month, self.start_date.day))
            except ParserError as err:
                raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        return end_date.year - self.start_date.year - ((end_date.month, end_date.day) < (self.start_date.month, self.start_date.day))
    
    # get months on a specific date
    def monthsOnDate(self,end_date= date.today()):
        try:
            end_date=end_date if isinstance(end_date,date) else parse(end_date)
        except ParserError as err:
                raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        
        years2months=(end_date.year - self.start_date.year)*12
        months2months=end_date.month-self.start_date.month
        return years2months+months2months
    
    #get days 
    @property
    def days(self):
        return (self.end_date-self.start_date).days
    
    def daysOnDate(self,end_date=date.today()):
        if not isinstance(end_date,date):
            try:
                end_date=parse(end_date).date()
                return (end_date- self.start_date).days
            except ParserError as err:
                raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        return (end_date- self.start_date).days

def main():
    d=Duration('2020-01-01','2022-01-02')
    print(d.days)

if __name__=="__main__":
    main()
    
    



from typing import List,Optional
from model.experience.ecdata import Employment
from model.common.employmenthistory import EmploymentHistory
from datetime import date
from source.excel import Excel
from model.common.id import ID,IDs
from model.common.person import Person
from model.common.wordmaker import WordMaker
import os

class EmploymentModel():
    
    #使用excels keyword参数是为了配合其他模块，用excel文件名列表输入。
    def __init__(self,excels=None): # excel includes data
        self.data_excel=excels[0]
    
    def getData(self):
        e = Excel(self.data_excel)
        #Get company set
        try:
            company_set = set([emp['company'] for emp in e.dict['employment'] if emp['employment_certificate'][0].upper() == 'Y'])
        except KeyError as err:
            raise ValueError(f'key error in employment sheet. Please check "company","employment","employment_certificate"')
        # pick a company for employment certificate generation
        companies = list(company_set)
        [print(index, company) for index, company in enumerate(companies)]
        which_one = input("Please input the number of the employer: ")
        company = companies[int(which_one)]

        # get validated data
        m_ec = []
        is_current=False
        works = [emp for emp in e.dict['employment'] if emp['company'] == company]
        
        for i, employment in enumerate(works):
            if not employment["end_date"]: is_current=True
            m_ec.append(Employment(**employment))
        #get summary info
        history=EmploymentHistory(m_ec)
        start_date=history.initial_start_date
        end_date=history.final_end_date
        position_number=history.position_number_say
        # person info
        personal=Person(**e.dict['personal'])
        iddds=e.dict['personid']
        id_list=[ID(**id) for id in iddds]
        ids=IDs(id_list)

        # get company brief, and hr rep's information from the list of positions in one company. The rep info could be in any period, but only get one of them.
        def getContent(field):
            for w in m_ec:
                value = getattr(w,field)
                if value:
                    return value
            return None

        context={
            # 'company_brief':company_brief,
            'company_brief':getContent('company_brief'),
            'hr_rep_name':getContent('fullname_of_certificate_provider'),
            'hr_rep_position':getContent('position_of_certificate_provider'),
            'hr_rep_department':getContent('department_of_certificate_provider'),
            'hr_rep_phone':getContent('phone_of_certificate_provider'),
            'hr_rep_email':getContent('email_of_certificate_provider'),
            "passport":ids.passport.number,
            "work":m_ec,
            'personal':personal,
            'more_than_1':len(m_ec)>1,
            'position_number':position_number,
            'is_current':is_current,
            'start_date':start_date.strftime('%b %Y'),
            'end_date':end_date
            }
        
        return context
    
    
        

class EmploymentModelDocxAdapater():
    """This is an adapater to bridging resume model data and docx data
    """
    def __init__(self,emp_obj: EmploymentModel):
        # get original resume obj, which will be used to generate some value based on it's object methods. 
        self.emp_obj=emp_obj
    
    def make(self,output_docx):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/word/employment_certificate.docx"            
        wm=WordMaker(template_path,self.emp_obj.getData(),output_docx)
        wm.make()
    



# Here define all errors and prompts in different language
error_dict={
    'none is not an allowed value':"不允许为空",
    'invalid date format':'日期格式错误，必须是这种格式： 2022-01-31',
    'file':'文件中的',
    'sheet':'表',
    'the_column':'该列',
    'has error':"有错误",
    'error':'错误',
    'detail':"详细信息"
}



from pydantic import ValidationError
from termcolor import colored
import json

class Validation(object):
    def __init__(self,validator_model,excel_name,sheet) -> None:
        self.validate_pair=validator_model.validate_pair
        self.excel_name=excel_name
        self.sheet_name=sheet.name
        self.sheet=sheet
        self.validated_data=None
    @property
    def json(self):
        return json.dumps(self.validated_data,indent=3,default=str)
        
class SheetValidation(Validation):
    """SheetValidation validate a sheet object, and return validated data from validated_data dict

    Args:
        validator_model: A pydandic model, in which must have a validate_pair dict used for pointing the processing class 
        excel_name: an excel file
        sheet: a sheet object (SheetDict object)
    """
    def __init__(self,validator_model,excel_name,sheet):
        super().__init__(validator_model,excel_name,sheet)
        self.validated_data={}
        self.validate()
        
    def interpretError(self,e:ValidationError):
        print(colored(f"Erros in file {self.excel_name}-> info-{self.sheet_name} sheet",'red'))
        for err in e.errors():
            for variable in err['loc']:
                info_node=self.sheet.data.get(variable)
                if info_node !=None: 
                    print(colored(f"{info_node.description}",'green'),err['msg'],err['type'].replace('.',': '))
            
    def validate(self):
        try:
            # get data from sheet object
            data={variable: info_node_obj.value for variable, info_node_obj in self.sheet.data.items()}
            validatedData=(self.validate_pair[self.sheet_name](**data))
            self.validated_data = validatedData.dict()
        except ValidationError as e:
            self.interpretError(e)
    

class TableValidation(Validation):
    """TableValidation validate a table object, and return validated data from validated_data list

    Args:
        validator_model: A pydandic model, in which must have a validate_pair dict used for pointing the processing class 
        excel_name: an excel file
        sheet: a table object (TableDict object)
    """
    def __init__(self,validator_model,excel_name,sheet) -> None:
        super().__init__(validator_model,excel_name,sheet)
        self.validated_data=[]
        self.row_index=0
        self.validate()
            
    def interpretError(self,e:ValidationError):
        print(colored(f"Erros in file {self.excel_name}-> table-{self.sheet_name} sheet",'red'))
        for err in e.errors():
            for variable in err['loc']:
                index=self.sheet.column_variables.index(variable)
                variable_title=self.sheet.column_titles[index]
                print(colored(f"Row {self.row_index+4}: ",'red'),colored(f"{variable_title}",'green'),err['msg'],err['type'].replace('.',': '))
    
    def validate(self):
        try:
            for dataItem in self.sheet.data:
                validatedData=(self.validate_pair[self.sheet_name](**dataItem.__dict__))
                self.validated_data.append(validatedData.dict())
                self.row_index+=1
        except ValidationError as e:
            self.interpretError(e)



# before '_' is sheet name
# variable as key, xpath of xml as value.
data_model = {
    'trcase': {
        'service_in': 'Page1/PersonalDetails/ServiceIn/ServiceIn',
        'same_as_cor': "Page1/PersonalDetails/SameAsCORIndicator",
        "applying_country": "Page1/PersonalDetails/CountryWhereApplying/Row2/Country",
        "applying_stauts": "Page1/PersonalDetails/CountryWhereApplying/Row2/Status",
        "applying_start_date": "Page1/PersonalDetails/CountryWhereApplying/Row2/FromDate",
        "applying_end_date": "Page1/PersonalDetails/CountryWhereApplying/Row2/ToDate",
        "consent_of_info_release": "Page4/Consent0/Choice",
        "submission_date": "Page4/Consent0/C1CertificateIssueDate"
    },
    'personal': {
        'uci': "Page1/PersonalDetails/UCIClientID",
        'last_name': "Page1/PersonalDetails/Name/FamilyName",
        'first_name': "Page1/PersonalDetails/Name/GivenName",
        'has_alias_name': "Page1/PersonalDetails/AliasName/AliasNameIndicator/AliasNameIndicator",
        'used_last_name': "Page1/PersonalDetails/AliasName/AliasFamilyName",
        'used_first_name': "Page1/PersonalDetails/AliasName/AliasGivenName",
        'sex': "Page1/PersonalDetails/Sex/Sex",
        'dob_year': "Page1/PersonalDetails/DOBYear",
        'dob_month': "Page1/PersonalDetails/DOBMonth",
        'dob_day': "Page1/PersonalDetails/DOBDay",
        'place_of_birth': "Page1/PersonalDetails/PlaceBirthCity",
        'country_of_birth': "Page1/PersonalDetails/PlaceBirthCountry",
        'citizen': "Page1/PersonalDetails/Citizenship/Citizenship",
        "native_language": "Page2/MaritalStatus/SectionA/Languages/languages/nativeLang/nativeLang",
        "english_french": "Page2/MaritalStatus/SectionA/Languages/languages/ableToCommunicate/ableToCommunicate",
        "which_one_better": "Page2/MaritalStatus/SectionA/Languages/languages/lov",
        "language_test": "Page2/MaritalStatus/SectionA/Languages/LanguageTest",
        "email": "Page3/FaxEmail/Email"
    },
    # country of residence: current and past for over 6 months. table model
    'cor': {
        'current_cor_country': "Page1/PersonalDetails/CurrentCOR/Row2/Country",
        'current_cor_status': "Page1/PersonalDetails/CurrentCOR/Row2/Status",
        'current_cor_other': "Page1/PersonalDetails/CurrentCOR/Row2/Other",
        'current_cor_start_date': "Page1/PersonalDetails/CurrentCOR/Row2/FromDate",
        'current_cor_end_date': "Page1/PersonalDetails/CurrentCOR/Row2/ToDate",
        "previous_cor_country1": "Page1/PersonalDetails/PreviousCOR/Row2/Country",
        "previous_cor_status1": "Page1/PersonalDetails/PreviousCOR/Row2/Status",
        "previous_cor_other1": "Page1/PersonalDetails/PreviousCOR/Row2/Other",
        "previous_cor_start_date1": "Page1/PersonalDetails/PreviousCOR/Row2/FromDate",
        "previous_cor_end_date1": "Page1/PersonalDetails/PreviousCOR/Row2/ToDate",
        "previous_cor_country2": "Page1/PersonalDetails/PreviousCOR/Row3/Country",
        "previous_cor_status2": "Page1/PersonalDetails/PreviousCOR/Row3/Status",
        "previous_cor_other2": "Page1/PersonalDetails/PreviousCOR/Row3/Other",
        "previous_cor_start_date2": "Page1/PersonalDetails/PreviousCOR/Row3/FromDate",
        "previous_cor_end_date2": "Page1/PersonalDetails/PreviousCOR/Row3/ToDate"
    },

    # marriage
    'marriage': {
        "marital_status": "Page1/MaritalStatus/SectionA/MaritalStatus",
        "married_date": "Page1/MaritalStatus/SectionA/DateOfMarriage",
        "sp_last_name": "Page1/MaritalStatus/SectionA/FamilyName",
        "sp_first_name": "Page1/MaritalStatus/SectionA/GivenName",
        "previous_married": "Page2/MaritalStatus/SectionA/PrevMarriedIndicator",
        "pre_sp_last_name": "Page2/MaritalStatus/SectionA/PMFamilyName",
        "pre_sp_first_name": "Page2/MaritalStatus/SectionA/GivenName/PMGivenName",
        "pre_relationship_type": "Page2/MaritalStatus/SectionA/TypeOfRelationship",
        "pre_sp_dob_year": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBYear",
        "pre_sp_dob_month": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBMonth",
        "pre_sp_dob_day": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBDay",
        "pre_start_date": "Page2/MaritalStatus/SectionA/FromDate",
        "pre_end_date": "Page2/MaritalStatus/SectionA/ToDate/ToDate"
    },

    'passport': {
        "number": "Page2/MaritalStatus/SectionA/Passport/PassportNum/PassportNum",
        "country": "Page2/MaritalStatus/SectionA/Passport/CountryofIssue/CountryofIssue",
        "issue_date": "Page2/MaritalStatus/SectionA/Passport/IssueDate/IssueDate",
        "expiry_date": "Page2/MaritalStatus/SectionA/Passport/ExpiryDate"
    },
    'national_id': {
        "has_id": "Page2/natID/q1/natIDIndicator",
        "number": "Page2/natID/natIDdocs/DocNum/DocNum",
        "country": "Page2/natID/natIDdocs/CountryofIssue/CountryofIssue",
        "issue_date": "Page2/natID/natIDdocs/IssueDate/IssueDate",
        "expiry_date": "Page2/natID/natIDdocs/ExpiryDate"
    },
    'us_pr': {
        "has_id": "Page2/USCard/q1/usCardIndicator",
        "number": "Page2/USCard/usCarddocs/DocNum/DocNum",
        "expiry_date": "Page2/USCard/usCarddocs/ExpiryDate"
    },

    'mailing_address': {
        'po_box': "Page2/ContactInformation/contact/AddressRow1/POBox/POBox",
        'unit': "Page2/ContactInformation/contact/AddressRow1/Apt/AptUnit",
        'street_number': "Page2/ContactInformation/contact/AddressRow1/StreetNum/StreetNum",
        'street_name': "Page2/ContactInformation/contact/AddressRow1/Streetname/Streetname",
        'city': "Page2/ContactInformation/contact/AddressRow2/CityTow/CityTown",
        'country': "Page2/ContactInformation/contact/AddressRow2/Country/Country",
        'province': "Page2/ContactInformation/contact/AddressRow2/ProvinceState/ProvinceState",
        'post_code': "Page2/ContactInformation/contact/AddressRow2/PostalCode/PostalCode",
        'district': "Page2/ContactInformation/contact/AddressRow2/District"
    },

    'residential_address': {
        'same_as_mailing': "Page2/ContactInformation/contact/SameAsMailingIndicator",
        'unit': "Page2/ContactInformation/contact/ResidentialAddressRow1/AptUnit/AptUnit",
        'street_number': "Page2/ContactInformation/contact/ResidentialAddressRow1/StreetNum/StreetNum",
        'street_name': "Page2/ContactInformation/contact/ResidentialAddressRow1/StreetName/Streetname",
        'city': "Page2/ContactInformation/contact/ResidentialAddressRow1/CityTown/CityTown",
        'country': "Page2/ContactInformation/contact/ResidentialAddressRow2/Country/Country",
        'province': "Page2/ContactInformation/contact/ResidentialAddressRow2/ProvinceState/ProvinceState",
        'post_code': "Page2/ContactInformation/contact/ResidentialAddressRow2/PostalCode/PostalCode",
        'district': "Page2/ContactInformation/contact/ResidentialAddressRow2/District"
    },

    'phone': {
        "variable_type": "Page2/ContactInformation/contact/PhoneNumbers/Phone/Type",
        "canada_us": "Page2/ContactInformation/contact/PhoneNumbers/Phone/CanadaUS",
        "other": "Page2/ContactInformation/contact/PhoneNumbers/Phone/Other",
        'ext': "Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberExt",
        'country_code': "Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberCountry",
        'number': "Page2/ContactInformation/contact/PhoneNumbers/Phone/ActualNumber",
        'na_area': "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/AreaCode",
        'na_first_three': "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/FirstThree",
        'na_last_five': "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/LastFive",
        'int_number': "Page2/ContactInformation/contact/PhoneNumbers/Phone/IntlNumber/IntlNumber"
    },

    'altphone': {
        "variable_type": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Type",
        "canada_us": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/CanadaUS",
        "other": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Other",
        'ext': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberExt",
        'country_code': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberCountry",
        'number': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/ActualNumber",
        'na_area': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/AreaCode",
        'na_first_three': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/FirstThree",
        'na_last_five': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/LastFive",
        'int_number': "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/IntlNumber/IntlNumber"
    },

    'fax': {
        "canada_us": "Page3/FaxEmail/Phone/CanadaUS",
        "other": "Page3/FaxEmail/Phone/Other",
        'ext': "Page3/FaxEmail/Phone/NumberExt",
        'country_code': "Page3/FaxEmail/Phone/NumberCountry",
        'number': "Page3/FaxEmail/Phone/ActualNumber",
        'na_area': "Page3/FaxEmail/Phone/NANumber/AreaCode",
        'na_first_three': "Page3/FaxEmail/Phone/NANumber/FirstThree",
        'na_last_five': "Page3/FaxEmail/Phone/NANumber/LastFive",
        'int_number': "Page3/FaxEmail/Phone/IntlNumber/IntlNumber"
    },
    'wp': {
        "work_permit_type": "Page3/DetailsOfIntendedWork/DetailsOfWork/TypeofWork/WorkPermitType",
        "employer_name": "Page3/DetailsOfIntendedWork/DetailsOfWork/PurposeRow1/EmployerName/EmployerName",
        "employer_address": "Page3/DetailsOfIntendedWork/DetailsOfWork/PurposeRow1/Address/Address",
        "employment_province": "Page3/IntendedLocationInCanada/intendedLocation/ProvinceState/ProvinceState",
        "employment_city": "Page3/IntendedLocationInCanada/intendedLocation/CityTown/CityTown",
        "work_location_address": "Page3/IntendedLocationInCanada/intendedLocation/Address",
        "job_title": "Page3/DetailsOfWorkCont/details/jobTitle",
        "brief_duties": "Page3/DetailsOfWorkCont/details/posDesc",
        "duration_from": "Page3/DetailsOfWorkCont/details/HowLongStudy/FromDate",
        "duration_to": "Page3/DetailsOfWorkCont/details/HowLongStudy/ToDate",
        "lmia_num_or_offer_num": "Page3/DetailsOfWorkCont/details/LMO/LMO"
    },
    'education': {
        "start_date_year": "Page3/PageWrapper/Education/Edu_Row1/FromYear",
        "start_date_month": "Page3/PageWrapper/Education/Edu_Row1/FromMonth",
        "end_date_year": "Page3/PageWrapper/Education/Edu_Row1/ToYear",
        "end_date_month": "Page3/PageWrapper/Education/Edu_Row1/ToMonth",
        "school_name": "Page3/PageWrapper/Education/Edu_Row1/School",
        "field_of_study": "Page3/PageWrapper/Education/Edu_Row1/FieldOfStudy",
        "city": "Page3/PageWrapper/Education/Edu_Row1/CityTown",
        "country": "Page3/PageWrapper/Education/Edu_Row1/Country/Country",
        "province": "Page3/PageWrapper/Education/Edu_Row1/ProvState"
    },
    # occupation is a special dict, and will be handled seperately
    'occupation': {
        "start_date_year": "FromYear",
        "start_date_month": "FromMonth",
        "end_date_year": "ToYear",
        "end_date_month": "ToMonth",
        "job_title": "Occupation/Occupation",
        "company": "Employer",
        "city": "CityTown/CityTown",
        "province": "ProvState",
        "country": "Country/Country"
    },

    'trbackground': {
        "q1a": "Page4/BackgroundInfo/Choice",
        "q1b": "Page4/BackgroundInfo/Choice",
        "q1c": "Page4/BackgroundInfo/Details/MedicalDetails",
        "q2a": "Page4/BackgroundInfo2/VisaChoice1",
        "q2b": "Page4/BackgroundInfo2/VisaChoice2",
        "q2c": "Page4/BackgroundInfo2/VisaChoice3",
        "q2d": "Page4/BackgroundInfo2/Details/refusedDetails",
        "q3a": "Page4/PageWrapper/BackgroundInfo3/Choice",
        "q3b": "Page4/PageWrapper/BackgroundInfo3/militaryServiceDetails",
        "q4a": "Page4/PageWrapper/Military/Choice",
        "q4b": "Page4/PageWrapper/Military/militaryServiceDetails",
        "q5": "Page4/PageWrapper/Occupation/Choice",
        "q6": "Page4/PageWrapper/GovPosition/Choice"
    }
}

convert_model = {
    'country_pairs': {
        "trcase": ["applying_country"],
        "personal": ['country_of_birth', 'citizen'],
        'passport': ['country'],
        'national_id': ['country'],
        'education': ['country'],
        'cor': ['current_cor_country', 'previous_cor_country1', 'previous_cor_country2']
    },
    'canada_province_pairs': {
        'wp': ["employment_province"],
        'address': ['province'],
        'education': ['province'],
        'employment': ['province']
    },
    'language_pairs': {
        'personal': ['native_language']
    },
    'phone_type_pairs': {
        'phone': ["variable_type"],
        'altphone': ["variable_type"]
    },
    'tr_canada_status_pairs': {
        'cor': ['current_cor_status', 'previous_cor_status1', 'previous_cor_status2']
    },
    'tr_marital_status_pairs': {
        'marriage': ['marital_status','pre_relationship_type']
    }
}

remove_model = {
    'personal': ['dob_year', 'dob_month', 'dob_day'],
    'marriage': ['pre_sp_dob_year', 'pre_sp_dob_month', 'pre_sp_dob_day']
}



from pydantic import BaseModel
from typing import Optional,List
from collections import namedtuple
from model.common.mixins import EqMixin

class Phone(BaseModel,EqMixin):
    variable_type:str
    display_type:str
    country_code:Optional[int]
    number:Optional[int]
    ext:Optional[int]
    

    class Config:
        anystr_lower=True 
    
    @property
    def international_format_full(self):
        return '+'+str(self.country_code)+" "+str(self.number)
    
    @property
    def isCanadaUs(self):
        return self.country_code and self.country_code==1
    
    @property
    def is_fax(self):
        return self.v_type=='fax'
    
    @property
    def NA_format_number(self):
        if len(str(self.number))!=10:
            raise ValueError(f"{self.number} is not a valid North America phone number")
        return '('+str(self.number)[:3]+') '+str(self.number)[3:6]+'-'+str(self.number)[6:]
    
        
    # Return a list including area code, first part and second part of number
    @property
    def NA_format_list(self):
        if len(str(self.number))!=10:
            raise ValueError(f"{self.number} is not a valid North America phone number")
        return [str(self.number)[:3],str(self.number)[3:6],str(self.number)[6:]]
    
    # Return a list including area code, first part and second part of number
    @property
    def NA_format_namedtuple(self):
        if len(str(self.number))!=10:
            raise ValueError(f"{self.number} is not a valid North America phone number")
        NA_Number=namedtuple('NA_Number',['area_code','part1','part2'])
        number=NA_Number(str(self.number)[:3],str(self.number)[3:6],str(self.number)[6:])
        return number
    
    def __str__(self):
        return self.international_format_full

class Phones(object):
    def __init__(self, phone_list: List[Phone]) -> None:
        self.phonees=phone_list
    
    def _specific_phone(self,v_type):
        phone=[phone for phone in self.phonees if phone.variable_type==v_type]
        return phone[0] if phone else None
    
    @property
    def cellular(self):
        return self._specific_phone('cellular')
    
    @property
    def residential(self):
        return self._specific_phone('residential')
    
    @property
    def business(self):
        return self._specific_phone('business')
    
    @property
    def working(self):
        return self._specific_phone('working')
    
    @property
    def PreferredPhone(self):
        for Phone_type in ['cellular','residential','business','working']:
            phone=self._specific_phone(Phone_type)
            if phone: return phone
    
    def getPreferredPhoneFromList(self,type_list):
        for Phone_type in type_list:
            phone=self._specific_phone(Phone_type)
            if phone: return phone
    




from docxtpl import DocxTemplate
from termcolor import colored

class WordMaker():
    """This class makes word file based on template and filling in the data of context
    """    
    def __init__(self,template,context,output):
        self.document=DocxTemplate(template)
        self.context=context
        self.output=output
    
    def make(self):
        self.document.render(self.context)
        self.document.save(self.output) 
        print(colored(f'{self.output} has been saved...','green'))



from datetime import date
from pydantic import BaseModel,EmailStr,validator
from typing import Optional
from model.common.person import Person
from model.common.utils import normalize
from pydantic.class_validators import root_validator

class Personal(Person):
    used_last_name:Optional[str]
    used_first_name:Optional[str]
    uci:Optional[str]
    country_of_birth:str
    place_of_birth:str
    email:EmailStr
    
    native_language:str
    english_french:str
    which_one_better:Optional[str]
    language_test:bool
    
    _normalize_used_first_name=validator('used_first_name',allow_reuse=True,check_fields=False)(normalize)
    _normalize_used_last_name=validator('used_last_name',allow_reuse=True,check_fields=False)(normalize)
    
    @root_validator
    def checkAnswers(cls,values):
        questions=['english_french']
        explanations=['which_one_better']
        qas=dict(zip(questions,explanations))
        for k,v in qas.items():  
            if values.get(k) and not values.get(v):
                    raise ValueError(f"Since {k} is true, but you did not answer the question {v} in info-position sheet")
        return values
    

    
class Marriage(BaseModel):
    marital_status:str
    married_date:Optional[date]
    sp_last_name:Optional[str]
    sp_first_name:Optional[str]
    previous_married:bool
    pre_sp_last_name:Optional[str]
    pre_sp_first_name:Optional[str]
    pre_relationship_type:Optional[str]
    pre_sp_dob:Optional[date]
    pre_start_date:Optional[date]
    pre_end_date:Optional[date]
    
class PersonId(BaseModel):
    variable_type:str
    display_type:str
    number:Optional[str]
    country:Optional[str]
    issue_date:Optional[date]
    expiry_date:Optional[date]

# Not everyone has education. So, it's optional. Without start date and end date, the app will regard having no education
class Education(BaseModel):
    start_date:Optional[date]
    end_date:Optional[date]
    school_name:Optional[str]
    education_level:Optional[str]
    field_of_study:Optional[str]
    city:Optional[str]
    province:Optional[str]
    country:Optional[str]
    
# Not everyone has employment experience. So, it's optional. Without start date and end date, the app will regard having no post secondary education   
class Employment(BaseModel):
    start_date:Optional[date]
    end_date:Optional[date]
    job_title:Optional[str]
    company:Optional[str]
    city:Optional[str]
    province:Optional[str]
    country:Optional[str]
    
class Travel(BaseModel):
    start_date:Optional[date]
    end_date:Optional[date]
    length:Optional[int]
    destination:Optional[str]
    purpose:Optional[str]
    
class Family(BaseModel):
    last_name:str	
    first_name:str	
    native_last_name:str	
    native_first_name:str
    marital_status:str	
    date_of_birth:Optional[date]	
    # place_of_birth:str	
    birth_country:str	
    # country_of_citizenship:str
    address:str
    occupation:str	
    relationship:str	
    # email:Optional[EmailStr]	
    # date_of_death:Optional[date]	
    # place_of_death:Optional[str]	
    accompany_to_canada:bool

# Countries of Residence    
class COR(BaseModel):
    start_date:date
    end_date:Optional[date]
    country:str
    status:str
    


from datetime import date
from pydantic import BaseModel
from typing import Optional
from model.common.person import Person

class PrCase(BaseModel):
    applying_country:str
    applying_stauts:str
    applying_start_date:date
    applying_end_date:date
    
# PR education is different from TR
class Education(BaseModel):
    highest_level:str
    years_of_education:int
    current_occupation:str
    intended_occupation:str
        
# Pr Personal has a little more properties than general person's 
class PrPersonal(Person):
    height:int
    eye_color:str
    


    
    


    



class XmlFiller():
    def __init__(self,template,context):
        self.template=template
        self.context=context
        
    def save(self,output):
        with open(self.template) as x:
            text=x.read()
            outtext=text.format(**self.context)
        with open(output,'w')as w:
            w.write(outtext)
            print(f' {output} Saved')



from dataclasses import dataclass
from datetime import date
from pydantic import BaseModel
from typing import Optional,List
from model.common.address import Address

class TrCase(BaseModel):
    service_in:str
    application_purpose:Optional[str] # 5257需要，1294 5 不需要
    same_as_cor:bool
    applying_country:str
    applying_stauts:str
    applying_start_date:date
    applying_end_date:Optional[date]
    consent_of_info_release:bool
    submission_date:Optional[date]

class TrCaseIn(BaseModel):
    service_in:str
    application_purpose:Optional[str] #TODO: consider
    original_entry_date:date
    original_entry_place:str
    original_purpose:str
    original_other_reason:Optional[str]
    most_recent_entry_date:date
    most_recent_entry_place:str
    doc_number:Optional[str]
    is_spouse_canadian:bool
    consent_of_info_release:bool
    submission_date:Optional[date]
    
class Visa(BaseModel):
    visit_purpose:str
    duration_from:date
    duration_to:date
    funds_available:int
    name1:Optional[str]
    relationship1:Optional[str]
    address1:Optional[str]
    name2:Optional[str]
    relationship:Optional[str]
    address2:Optional[str]

class Sp(BaseModel):
    school_name:str
    study_level:str
    study_field:str
    province:str
    city:str
    address:str
    dli:str
    student_id:str
    duration_from:date
    duration_to:date
    tuition_cost:Optional[str]
    room_cost:Optional[str]
    other_cost:Optional[str]
    fund_available:str
    paid_person:str
    other:Optional[str]    


class Wp(BaseModel):
    work_permit_type:str
    employer_name:Optional[str]
    employer_address:Optional[str]
    work_province:Optional[str]
    work_city:Optional[str]
    work_address:Optional[str]
    job_title:Optional[str]
    brief_duties:Optional[str]
    duration_from:date
    duration_to:date
    lmia_num_or_offer_num:Optional[str]
    
class VrInCanada(BaseModel):
    application_purpose:str
    visit_purpose:str
    duration_from:date
    duration_to:date
    funds_available:int
    name1:Optional[str]
    relationship1:Optional[str]
    address1:Optional[str]
    name2:Optional[str]
    relationship:Optional[str]
    address2:Optional[str]

class SpInCanada(BaseModel):
    application_purpose:str
    apply_work_permit:bool
    work_permit_type:str
    caq_number:Optional[str]
    expiry_date:Optional[date]
    school_name:str
    study_level:str
    study_field:str
    province:str
    city:str
    address:str
    dli:str
    student_id:str
    duration_from:date
    duration_to:date
    tuition_cost:Optional[str]
    room_cost:Optional[str]
    other_cost:Optional[str]
    fund_available:str
    paid_person:str
    other:Optional[str]    
    
class WpInCanada(BaseModel):
    application_purpose:str
    caq_number:Optional[str]
    expiry_date:Optional[date]
    work_permit_type:str
    employer_name:Optional[str]
    employer_address:Optional[str]
    work_province:Optional[str]
    work_city:Optional[str]
    work_address:Optional[str]
    job_title:Optional[str]
    brief_duties:Optional[str]
    duration_from:date
    duration_to:date
    lmia_num_or_offer_num:Optional[str]

class TrBackground(BaseModel):
    q1a:bool
    q1b:bool
    q1c:Optional[str]
    q2a:bool
    q2b:bool
    q2c:bool
    q2d:Optional[str]
    q3a:bool
    q3b:Optional[str]
    q4a:bool
    q4b:Optional[str]
    q5:bool
    q6:bool  
 


from pydantic import BaseModel,validator
from datetime import date,timedelta
from typing import Optional,List
from model.common.mixins import DurationMixin
from model.common.utils import makeList,Duration
from dateutil.parser import parse,ParserError
from model.common.employmentbase import EmploymentBase

class EmploymentHistory():
    def __init__(self,employment_list: List[EmploymentBase]):
        self.employment_list=employment_list
        for emp in employment_list:
            if not isinstance(emp,EmploymentBase):
                raise ValueError(f'{employment_list} is not a list of EmploymentBase objects')

    @property    
    def initial_start_date(self):
        return min([ emp.start_date for emp in self.employment_list])
    
    @property    
    def final_end_date(self):
        end_dates=[]
        for emp in self.employment_list:
            if emp.is_current: return "Present"
            end_dates.append(emp.end_date)
        print(end_dates)
        return max(end_dates)
    
    @property
    def position_number(self):
        return len(set([emp.job_title for emp in self.employment_list]))
    
    @property
    def position_number_say(self):
        number= len(set([emp.job_title for emp in self.employment_list]))
        return str(number)+" position" if number<=1 else str(number)+" positions"



from dataclasses import dataclass
from typing import List
from model.common.commonmodel import CommonModel
from model.bcpnp.data import Contact, General,JobOffer, PersonalAssess, Bcpnp
from model.common.jobposition import Position
from model.common.person import Person
from model.common.advertisement import Advertisement, Advertisements, InterviewRecord, RecruitmentSummary, InterviewRecords
from model.common.contact import Contacts
from model.common.wordmaker import WordMaker
import os


class Personal(Person):
    def __str__(self):
        return self.full_name


class RecommendationLetterModel(CommonModel):
    bcpnp: Bcpnp
    general: General
    contact: List[Contact]
    position: Position
    personal: Personal
    joboffer: JobOffer
    personalassess: PersonalAssess
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    recruitmentsummary: RecruitmentSummary

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            import os
            path = os.path.abspath(os.path.join(
                os.path.dirname(__file__), os.path.pardir))
            excels = [
                path+'/template/excel/er.xlsx',
                path+"/template/excel/pa.xlsx",
                path+"/template/excel/recruitment.xlsx",
                path+"/template/excel/bcpnp.xlsx",
                path+"/template/excel/rep.xlsx"
            ]
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    'You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels, output_excel_file, globals())

    @property
    def selected_contact(self):
        contacts = Contacts(self.contact)
        return contacts.preferredContact

    @property
    def summary(self):
        return InterviewRecords(self.interviewrecord)

    @property
    def advertisements(self):
        return Advertisements(self.advertisement)

    @property
    def person(self):
        return {
            "full_name": self.personal.full_name,
            "attributive": self.personal.attributive,
            "object": self.personal.object,
            "subject": self.personal.subject,
            'short_name': self.personal.short_name
        }


class RecommendationLetterDocxAdaptor():
    def __init__(self, recommdation_letter_obj: RecommendationLetterModel):
        self.recommdation_letter_obj = recommdation_letter_obj

    def re_generate_dict(self):
        summary_info = {
            'resume_num': self.recommdation_letter_obj.summary.resume_num,
            'canadian_num': self.recommdation_letter_obj.summary.canadian_num,
            'unknown_num': self.recommdation_letter_obj.summary.unknown_num,
            "foreigner_num": self.recommdation_letter_obj.summary.foreigner_num,
            "total_canadian": self.recommdation_letter_obj.summary.total_canadian,
            "total_interviewed_canadians": self.recommdation_letter_obj.summary.total_interviewed_canadians,
            "canadian_records": self.recommdation_letter_obj.summary.canadian_records,
            'contact': self.recommdation_letter_obj.selected_contact,
            "advertisement": self.recommdation_letter_obj.advertisements,
            'personal': self.recommdation_letter_obj.person,
            "work_start_date": self.recommdation_letter_obj.joboffer.start_date_say,
            'joboffer_date': self.recommdation_letter_obj.joboffer.date_of_offer
        }
        return {**self.recommdation_letter_obj.dict(), **summary_info}

    def make(self, output_docx):
        path = os.path.abspath(os.path.join(
            os.path.dirname(__file__), os.path.pardir))
        template_path = path+"/template/word/bcpnp_employer_recommendation_letter.docx"
        wm = WordMaker(template_path, self.re_generate_dict(), output_docx)
        wm.make()



from typing import List
from model.common.commonmodel import CommonModel
from model.bcpnp.data import JobOffer
from model.common.person import Person
from model.common.wordmaker import WordMaker
import os

class Personal(Person):
    def __str__(self):
        return self.full_name
        
class JobDescriptionModel(CommonModel):
    personal:Personal
    joboffer:JobOffer
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+"/template/excel/er.xlsx",
                path+"/template/excel/pa.xlsx"
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())


class JobDescriptionDocxAdaptor():
    def __init__(self,jobdescription_obj:JobDescriptionModel):
        self.jobdescription_obj=jobdescription_obj

    def make(self,output_docx):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/word/bcpnp_jobdescription.docx"            
        wm=WordMaker(template_path,self.jobdescription_obj,output_docx)
        wm.make()



from functools import reduce
from pydantic import BaseModel,ValidationError
from termcolor import colored
from source.excel import Excel
from model.common.errordict import error_dict

class CommonModel(BaseModel):
    # every succesor class has its own properties
    class Config:
        arbitrary_types_allowed = True
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. 
    # output_excel_file is finename for generating new excel. global_dict is the dict generated by globals() from successor's class
    def __init__(self,excels,output_excel_file,global_dict,generate_excel=True):
        # add up all input excels and return the combined one
        *original_excel_objs,new_excel_obj=self._addupExcels(excels) # the last one is new_excel_obj, before this are all original_objs 
        if output_excel_file:
            data=self._generateExcel(global_dict,new_excel_obj,output_excel_file)
            if generate_excel: self.generateExcel(**data)
        else:
            #get the dict as input kwargs. Here globals() is very important. Without it, the KwargsMix cannot find required classes
            kwargs=self._getInput(global_dict,original_excel_objs,new_excel_obj)
            # use pre-assigned referrence value (kwargs) to initialize the model
            try:
                super().__init__(**kwargs)
            except ValidationError as e:
                exit(0)    # since in _getinput has capatured exceptions, here we do nothing.

    # get combined excel throgh add up all input excels. 
    def _addupExcels(self,excels):
        excel_objs=list(map(lambda x:Excel(x), excels))
        return [*excel_objs,reduce(lambda a,b:a+b, excel_objs)]
    
    """
    Prerequisite of below methods
    1. excel_obj is the one that all input excels combined, no sheet has same name
    2. Successor's model is constructed with all sheets/tables, no specific single varialbe
    3. Only InfoSheet and TableList two data structure in the successor's model properties
    """
    def _getInput(self,global_dict,original_excel_objs,new_excel_obj):

        input_kwargs={}
        # According to the prerequisite described above, all the referrenced scheme are defined in 'definitions'. 
        # key is Model name, and after lower(), it is also excel sheet name and the property defined in this class. They are same. This is the requirement.
        for k in self.schema()['definitions'].keys(): 
            # globals() returns a dict inlcuding all global classes, methods, ....
            klass=global_dict.get(k)  
            # use k.lower (it's a sheet name here) to get sheet data
            sheet_data=new_excel_obj.dict.get(k.lower(),None)   
            # only two types, one is list of model objects,another is model object
            if not isinstance(sheet_data,list):
                # assign the input kwargs. Here k.lower() is class's property. klass is the model, on which the app create an obj
                try:
                    input_kwargs[k.lower()]=klass(**sheet_data)
                except ValidationError as e:
                    sheet_name='info-'+k.lower() # sheet name
                    excel_obj=[ excel_obj for excel_obj in original_excel_objs if sheet_name in excel_obj.sheet_names][0]
                    self.interpretError(excel_obj,sheet_name,e)
            else:
                try:
                    input_kwargs[k.lower()]=[klass(**d) for d in sheet_data]
                except ValidationError as e:
                    table_name='table-'+k.lower() # sheet name
                    excel_obj=[ excel_obj for excel_obj in original_excel_objs if table_name in excel_obj.table_names][0]
                    self.interpretError(excel_obj,table_name,e)
        return input_kwargs

    # generate a new excel according to the program class's properties    
    def _generateExcel(self,global_dict,new_excel_obj,excel_file):    
        sheets=[]
        tables=[]
        # successo's referrenced class is retored in schema's definition fields with dict format.
        # 获得每一sheet对应class，并获得它的properties list
        classes=self.schema()['definitions'].keys()
        for k in classes: 
            # global_dict is a dict inlcuding all global classes, methods, ....
            klass=global_dict.get(k) # k here is actually a sheet/table name, and klass is the class related to sheet. 
            variables=list(klass.__fields__.keys()) # get all variables in the sheet/table
            sheet_table_name=k.lower() # k.lower is the sheet / table name according to convention
            if 'info-'+sheet_table_name in new_excel_obj.sheet_names:
                sheets.append(new_excel_obj.getSheet(sheet_table_name,variables)) # get sheet object 
            if 'table-'+sheet_table_name in new_excel_obj.table_names:
                tables.append(new_excel_obj.getTable(sheet_table_name,variables)) # get table object
        return {'new_excel_obj':new_excel_obj,'excel_file':excel_file,'sheets':sheets,'tables':tables}
    
    def generateExcel(self,**data):        
        # 根据最终更新了variables的sheets和talbes，生成新的excel文件
        data['new_excel_obj'].makeExcel(data['excel_file'],sheets=data['sheets'],tables=data['tables'])

    # def interpretError(self,excel_obj,sheet_name,e:ValidationError):
    #     excel_name=excel_obj.excel_name.split('/')[-1] # only get the excel file name and ignore the path
    #     print(colored(f"{excel_name}{error_dict.get('file')}{sheet_name}{error_dict.get('sheet')}{error_dict.get('has error')}，{error_dict.get('detail')}: ",'red'))
    #     for err in e.errors():
    #         for variable in err['loc']:
    #             if sheet_name.startswith('info-'):
    #                 sheet=excel_obj.getSheet(sheet_name[5:]) # remove info-
    #                 info_node=sheet.data.get(variable)
    #                 if info_node !=None: 
    #                     print(colored(f"{info_node.description}",'red'),error_dict.get(err['msg']),error_dict.get('has error'))
    #             if sheet_name.startswith('table-'):
    #                 table=excel_obj.getTable(sheet_name[6:])
    #                 title=table.column_titles.get(variable)
    #                 print(colored(f"{title}",'red'),error_dict.get('the_column'),error_dict.get('has error'),error_dict.get(err['msg'],error_dict.get('error')))        
    #     print(e.json)        
    
    def interpretError(self,excel_obj,sheet_name,e:ValidationError):
        excel_name=excel_obj.excel_name.split('/')[-1] # only get the excel file name and ignore the path
        print(colored(f"{excel_name}{error_dict.get('file')}{sheet_name}{error_dict.get('sheet')}{error_dict.get('has error')}，{error_dict.get('detail')}: ",'red'))
        for err in e.errors():
            for variable in err['loc']:
                if sheet_name.startswith('info-'):
                    sheet=excel_obj.getSheet(sheet_name[5:]) # remove info-
                    info_node=sheet.data.get(variable)
                    if info_node !=None: 
                        print(colored(f"{info_node.description}",'red'),err['msg'])
                if sheet_name.startswith('table-'):
                    table=excel_obj.getTable(sheet_name[6:])
                    title=table.column_titles.get(variable)
                    print(colored(f"{title}",'red'),'the_column has error',err['msg'])
        print(e)        



from functools import reduce
from pydantic import BaseModel,validator
from datetime import date
from typing import Optional,List
from model.common.mixins import DurationMixin
from model.common.utils import makeList,normalize

"""
with properties of start_from and end_to got from mixin
"""
class EducationBase(BaseModel,DurationMixin):
    start_date: date
    end_date:Optional[date]
    school_name: str
    education_level:str
    field_of_study:str
    
    _normalize_school_name=validator('school_name',allow_reuse=True,check_fields=False)(normalize)
    _normalize_field_of_study=validator('field_of_study',allow_reuse=True,check_fields=False)(normalize)
    
    @validator("end_date")
    def endDateBigger(cls,end_date,values):
        start_date=values.get('start_date')
        the_date=end_date if end_date else date.today()
        if (the_date-start_date).days<=0:
            raise ValueError(f'End date {the_date} is earlier than start date {start_date}')
        end_date=end_date if end_date else "Present"
        return end_date

class EducationHistory():
    def __init__(self,edu_list:List[EducationBase]):
        self.edu_list=edu_list
    
    # total education years
    @property
    def years(self):
        return reduce(lambda a,b:a.lengthOfYears+b.lengthOfYears,self.edu_list)




# before '_' is sheet name
# variable as key, xpath of xml as value. 
data_model={
    'trcase':{
        'service_in':'Page1/PersonalDetails/ServiceIn/ServiceIn',
        "application_purpose":"Page1/PersonalDetails/VisaType/VisaType",
        'same_as_cor':"Page1/PersonalDetails/SameAsCORIndicator",
        "applying_country":"Page1/PersonalDetails/CountryWhereApplying/Row2/Country",
        "applying_stauts":"Page1/PersonalDetails/CountryWhereApplying/Row2/Status",
        "applying_start_date":"Page1/PersonalDetails/CountryWhereApplying/Row2/FromDate",
        "applying_end_date":"Page1/PersonalDetails/CountryWhereApplying/Row2/ToDate",
        "consent_of_info_release":"Page3/Signature/Consent0/Choice",
        "submission_date":"Page3/Signature/C1CertificateIssueDate"
    },
    "visa":{
        "visit_purpose":"Page3/DetailsOfVisit/PurposeRow1/PurposeOfVisit/PurposeOfVisit",
        "other":"Page3/DetailsOfVisit/PurposeRow1/Other/Other",
        "duration_from":"Page3/DetailsOfVisit/PurposeRow1/HowLongStay/FromDate",
        "duration_to":"Page3/DetailsOfVisit/PurposeRow1/HowLongStay/ToDate",
        "funds_available":"Page3/DetailsOfVisit/PurposeRow1/Funds/Funds",
        "name1":"Page3/DetailsOfVisit/Contacts_Row1/Name/Name",
        "relationship1":"Page3/DetailsOfVisit/Contacts_Row1/RelationshipToMe/RelationshipToMe",
        "address1":"Page3/DetailsOfVisit/Contacts_Row1/AddressInCanada/AddressInCanada",
        "name2":"Page3/Contacts_Row2/Name/Name",
        "relationship":"Page3/Contacts_Row2/Relationship/RelationshipToMe",
        "address2":"Page3/Contacts_Row2/AddressInCanada/AddressInCanada"
    },
    'personal':{
        'uci':"Page1/PersonalDetails/UCIClientID",
        'last_name':"Page1/PersonalDetails/Name/FamilyName",
        'first_name':"Page1/PersonalDetails/Name/GivenName",
        'has_alias_name':"Page1/PersonalDetails/AliasName/AliasNameIndicator/AliasNameIndicator",
        'used_last_name':"Page1/PersonalDetails/AliasName/AliasFamilyName",
        'used_first_name':"Page1/PersonalDetails/AliasName/AliasGivenName",
        'sex':"Page1/PersonalDetails/Sex/Sex",
        'dob_year':"Page1/PersonalDetails/DOBYear",
        'dob_month':"Page1/PersonalDetails/DOBMonth",
        'dob_day':"Page1/PersonalDetails/DOBDay",
        'place_of_birth':"Page1/PersonalDetails/PlaceBirthCity",
        'country_of_birth':"Page1/PersonalDetails/PlaceBirthCountry",
        'citizen':"Page1/PersonalDetails/Citizenship/Citizenship",
        "native_language":"Page2/MaritalStatus/SectionA/Languages/languages/nativeLang/nativeLang",
        "english_french":"Page2/MaritalStatus/SectionA/Languages/languages/ableToCommunicate/ableToCommunicate",
        "which_one_better":"Page2/MaritalStatus/SectionA/Languages/languages/lov",
        "language_test":"Page2/MaritalStatus/SectionA/Languages/LanguageTest",
        "email":"Page2/ContactInformation/contact/FaxEmail/Email"
    },
    #country of residence: current and past for over 6 months. table model
    'cor':{
        'current_cor_country':"Page1/PersonalDetails/CurrentCOR/Row2/Country",
        'current_cor_status':"Page1/PersonalDetails/CurrentCOR/Row2/Status",
        'current_cor_other':"Page1/PersonalDetails/CurrentCOR/Row2/Other",
        'current_cor_start_date':"Page1/PersonalDetails/CurrentCOR/Row2/FromDate",
        'current_cor_end_date':"Page1/PersonalDetails/CurrentCOR/Row2/ToDate",
        "previous_cor_country1":"Page1/PersonalDetails/PreviousCOR/Row2/Country",
        "previous_cor_status1":"Page1/PersonalDetails/PreviousCOR/Row2/Status",
        "previous_cor_other1":"Page1/PersonalDetails/PreviousCOR/Row2/Other",
        "previous_cor_start_date1":"Page1/PersonalDetails/PreviousCOR/Row2/FromDate",
        "previous_cor_end_date1":"Page1/PersonalDetails/PreviousCOR/Row2/ToDate",
        "previous_cor_country2":"Page1/PersonalDetails/PreviousCOR/Row3/Country",
        "previous_cor_status2":"Page1/PersonalDetails/PreviousCOR/Row3/Status",
        "previous_cor_other2":"Page1/PersonalDetails/PreviousCOR/Row3/Other",
        "previous_cor_start_date2":"Page1/PersonalDetails/PreviousCOR/Row3/FromDate",
        "previous_cor_end_date2":"Page1/PersonalDetails/PreviousCOR/Row3/ToDate"
    },

    # marriage
    'marriage':{
        "marital_status":"Page1/MaritalStatus/SectionA/MaritalStatus",
        "married_date":"Page1/MaritalStatus/SectionA/DateOfMarriage",
        "sp_last_name":"Page1/MaritalStatus/SectionA/FamilyName",
        "sp_first_name":"Page1/MaritalStatus/SectionA/GivenName",
        "previous_married":"Page2/MaritalStatus/SectionA/PrevMarriedIndicator",
        "pre_sp_last_name":"Page2/MaritalStatus/SectionA/PMFamilyName",
        "pre_sp_first_name":"Page2/MaritalStatus/SectionA/GivenName/PMGivenName",
        "pre_relationship_type":"Page2/MaritalStatus/SectionA/TypeOfRelationship",
        "pre_sp_dob_year":"Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBYear",
        "pre_sp_dob_month":"Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBMonth",
        "pre_sp_dob_day":"Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBDay",
        "pre_start_date":"Page2/MaritalStatus/SectionA/FromDate",
        "pre_end_date":"Page2/MaritalStatus/SectionA/ToDate/ToDate"
    },

    'passport':{
        "number":"Page2/MaritalStatus/SectionA/Passport/PassportNum/PassportNum",
        "country":"Page2/MaritalStatus/SectionA/Passport/CountryofIssue/CountryofIssue",
        "issue_date":"Page2/MaritalStatus/SectionA/Passport/IssueDate/IssueDate",
        "expiry_date":"Page2/MaritalStatus/SectionA/Passport/ExpiryDate"
    },
    'national_id':{
        "has_id":"Page2/natID/q1/natIDIndicator",
        "number":"Page2/natID/natIDdocs/DocNum/DocNum",
        "country":"Page2/natID/natIDdocs/CountryofIssue/CountryofIssue",
        "issue_date":"Page2/natID/natIDdocs/IssueDate/IssueDate",
        "expiry_date":"Page2/natID/natIDdocs/ExpiryDate"
    },
    'us_pr':{
        "has_id":"Page2/USCard/q1/usCardIndicator",
        "number":"Page2/USCard/usCarddocs/DocNum/DocNum",
        "expiry_date":"Page2/USCard/usCarddocs/ExpiryDate"
    },

    'mailing_address':{
        'po_box':"Page2/ContactInformation/contact/AddressRow1/POBox/POBox",
        'unit':"Page2/ContactInformation/contact/AddressRow1/Apt/AptUnit",
        'street_number':"Page2/ContactInformation/contact/AddressRow1/StreetNum/StreetNum",
        'street_name':"Page2/ContactInformation/contact/AddressRow1/Streetname/Streetname",
        'city':"Page2/ContactInformation/contact/AddressRow2/CityTow/CityTown",
        'country':"Page2/ContactInformation/contact/AddressRow2/Country/Country",
        'province':"Page2/ContactInformation/contact/AddressRow2/ProvinceState/ProvinceState",
        'post_code':"Page2/ContactInformation/contact/AddressRow2/PostalCode/PostalCode",
        'district':"Page2/ContactInformation/contact/AddressRow2/District"
    },

    'residential_address':{
        'same_as_mailing':"Page2/ContactInformation/contact/SameAsMailingIndicator",
        'unit':"Page2/ContactInformation/contact/ResidentialAddressRow1/AptUnit/AptUnit",
        'street_number':"Page2/ContactInformation/contact/ResidentialAddressRow1/StreetNum/StreetNum",
        'street_name':"Page2/ContactInformation/contact/ResidentialAddressRow1/StreetName/Streetname",
        'city':"Page2/ContactInformation/contact/ResidentialAddressRow1/CityTown/CityTown",
        'country':"Page2/ContactInformation/contact/ResidentialAddressRow2/Country/Country",
        'province':"Page2/ContactInformation/contact/ResidentialAddressRow2/ProvinceState/ProvinceState",
        'post_code':"Page2/ContactInformation/contact/ResidentialAddressRow2/PostalCode/PostalCode",
        'district':"Page2/ContactInformation/contact/ResidentialAddressRow2/District"
    },

    'phone':{
        "variable_type":"Page2/ContactInformation/contact/PhoneNumbers/Phone/Type",
        "canada_us":"Page2/ContactInformation/contact/PhoneNumbers/Phone/CanadaUS",
        "other":"Page2/ContactInformation/contact/PhoneNumbers/Phone/Other",
        'ext':"Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberExt",
        'country_code':"Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberCountry",
        'number':"Page2/ContactInformation/contact/PhoneNumbers/Phone/ActualNumber",
        'na_area':"Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/contact/PhoneNumbers/Phone/IntlNumber/IntlNumber"
    },

    'altphone':{
        "variable_type":"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Type",
        "canada_us":"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/CanadaUS",
        "other":"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Other",
        'ext':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberExt",
        'country_code':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberCountry",
        'number':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/ActualNumber",
        'na_area':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/contact/PhoneNumbers/AltPhone/IntlNumber/IntlNumber"
    },

    'fax':{
        "canada_us":"Page2/ContactInformation/contact/FaxEmail/Phone/CanadaUS",
        "other":"Page2/ContactInformation/contact/FaxEmail/Phone/Other",
        'ext':"Page2/ContactInformation/contact/FaxEmail/Phone/NumberExt",
        'country_code':"Page2/ContactInformation/contact/FaxEmail/Phone/NumberCountry",
        'number':"Page2/ContactInformation/contact/FaxEmail/Phone/ActualNumber",
        'na_area':"Page2/ContactInformation/contact/FaxEmail/Phone/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/contact/FaxEmail/Phone/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/contact/FaxEmail/Phone/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/contact/FaxEmail/Phone/IntlNumber/IntlNumber"
    },

    'education':{
        "start_date_year":"Page3/Education/Edu_Row1/FromYear",
        "start_date_month":"Page3/Education/Edu_Row1/FromMonth",
        "end_date_year":"Page3/Education/Edu_Row1/ToYear",
        "end_date_month":"Page3/Education/Edu_Row1/ToMonth",
        "school_name":"Page3/Education/Edu_Row1/School",
        "field_of_study":"Page3/Education/Edu_Row1/FieldOfStudy",
        "city":"Page3/Education/Edu_Row1/CityTown",
        "country":"Page3/Education/Edu_Row1/Country/Country",
        "province":"Page3/Education/Edu_Row1/ProvState"
    },
    # occupation is a special dict, and will be handled seperately
    'occupation':{
        "start_date_year":"FromYear",
        "start_date_month":"FromMonth",
        "end_date_year":"ToYear",
        "end_date_month":"ToMonth",
        "job_title":"Occupation/Occupation",
        "company":"Employer",
        "city":"CityTown/CityTown",
        "province":"ProvState",
        "country":"Country/Country"
    },

    'trbackground':{
        "q1a":"Page3/BackgroundInfo/Choice",
        "q1b":"Page3/BackgroundInfo/Choice",
        "q1c":"Page3/BackgroundInfo/Details/MedicalDetails",
        "q2a":"Page3/BackgroundInfo2/VisaChoice1",
        "q2b":"Page3/BackgroundInfo2/VisaChoice2",
        "q2c":"Page3/BackgroundInfo2/Details/VisaChoice3",
        "q2d":"Page3/BackgroundInfo2/Details/refusedDetails",
        "q3a":"Page3/BackgroundInfo3/Choice",
        "q3b":"Page3/PageWrapper/BackgroundInfo3/militaryServiceDetails",
        "q3a":"Page3/PageWrapper/Military/Choice",
        "q3b":"Page3/PageWrapper/Military/militaryServiceDetails",
        "q5":"Page3/PageWrapper/Occupation/Choice",
        "q6":"Page3/PageWrapper/GovPosition/Choice"
    }
}
convert_model = {
    "country_pairs": {
        "trcase": ["applying_country"],
        "personal": ['country_of_birth', 'citizen'],
        'passport': ['country'],
        'national_id': ['country'],
        'education': ['country'],
        'cor': ['current_cor_country', 'previous_cor_country1', 'previous_cor_country2']
    },
    'canada_province_pairs': {
        'visa': ["province"],
        'address': ['province'],
        'education': ['province'],
        'employment': ['province']
    },
    'language_pairs': {
        'personal': ['native_language']
    },
    'phone_type_pairs': {
        'phone': ["variable_type"],
        'altphone': ["variable_type"]
    },

    'tr_portal_study_level_pairs': {
        'sp': ["study_level"]
    },
    'tr_portal_study_field_pairs': {
        'sp': ["study_field"]
    },

    'tr_canada_status_pairs': {
        'cor': ['current_cor_status', 'previous_cor_status1', 'previous_cor_status2']
    },
    'tr_marital_status_pairs': {
        'marriage': ['marital_status', 'pre_relationship_type']
    }
}

remove_model = {
    'personal': ['dob_year', 'dob_month', 'dob_day'],
    'marriage': ['pre_sp_dob_year', 'pre_sp_dob_month', 'pre_sp_dob_day']
}


# before '_' is sheet name
# variable as key, xpath of xml as value. 
data_model={
    'education':{
        "name":"name",
        "education_level":"position",
        "field_of_study":"description"
    },
    'employment':{
        "company":"name",
        "job_title":"position",
        'duties':"description"
    },
    'travel':{
        "destination":"description",
        "purpose":"name",
    }
}



from pydantic import BaseModel
from typing import Optional
from pydantic.class_validators import root_validator

class Position(BaseModel):
    why_hire:str
    who_current_fill:str
    how_did_you_find:str
    how_when_offer:str
    worked_working:bool
    worked_working_details:Optional[str]
    has_same:bool
    lowest:Optional[float]
    highest:Optional[float]
    under_cba:bool
    which_union:Optional[str]
    lmia_refused:bool
    lmia_refused_reason:Optional[str]
    
    @root_validator
    def checkAnswers(cls,values):
        questions=['worked_working','under_cba','lmia_refused']
        explanations=['worked_working_details','which_union','lmia_refused_reason']
        qas=dict(zip(questions,explanations))
        for k,v in qas.items():  
            if values.get(k) and not values.get(v):
                    raise ValueError(f"Since {k} is true, but you did not answer the question {v} in info-position sheet")
        return values
    
    @root_validator
    def checkRange(cls,values):
        if values.get('has_same')==True:
            if not values['lowest'] or not values['highest']:
                raise ValueError(f'Since you indicated in info-position sheet has same employee in this position, you have to input lowest and highest wage.')
            elif values['lowest']>values['highest']:
                raise ValueError("range must have two number ,and the first is lower, the second is higher")
            else:
                return values 
        return values

    


from ast import Str
from asyncio import queues
from sqlite3 import Date
from pydantic import BaseModel,validator,EmailStr
from datetime import date
from typing import Optional,List,Union
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.contact import ContactBase
from model.common.utils import Duration
from model.common.address import Address
from pydantic.class_validators import root_validator
from model.common.utils import makeList

# bcpnp
class Bcpnp(BaseModel):
    has_applied_before:bool
    pre_file_no:Optional[str]
    account:str
    password:str
    submission_date:date
    case_stream:str
    q1:bool
    q1_explaination:Optional[str]
    q2:bool
    q2_explaination:Optional[str]
    q3:bool
    q3_explaination:Optional[str]
    q4:bool
    q4_file_number:Optional[str]
    q4_explaination:Optional[str]
    q5:bool
    q5_explaination:Optional[str]
    q6:bool
    q6_explaination:Optional[str]
    q7:bool
    q7_explaination:Optional[str]

    @root_validator
    def checkAnswers(cls,values):
        questions=['q1','q2','q3','q4','q5','q6','q7']
        explanations=[ q+'_explaination' for q in questions]
        qas=dict(zip(questions,explanations))
        qas['has_applied_before']="pre_file_no"
        for k,v in qas.items():  
            if values.get(k) and not values.get(v):
                    raise ValueError(f"Since {k} is true, but you did not answer the question {v} in info-bcpnp sheet")
        return values


# employer classes
class General(EmployerBase):
    company_intro:str
    business_intro:str
    company_more:Optional[str]
    recruit_email:EmailStr
    industry:str
    corporate_structure:str
    registration_number:str
    recruit_email:EmailStr
    ft_employee_number:int
    establish_date:date

#employer address
class ErAddress(Address):
    pass

class Contact(ContactBase):
    position:Optional[str]
    
#personal classes
class JobOffer(JobofferBase):
    offer_date:date
    is_working:bool
    work_start_date:Optional[date]
    supervisor_name:str
    supervisor_title:str
    other_language_required:bool
    reason_for_other:Optional[str]
    license_request:bool
    license_description:Optional[str]
    duties:list
    specific_edu_requirement:Optional[str]
    skill_experience_requirement:Optional[str]
    other_requirements:Optional[list]

    _str2bool_duties=validator('duties',allow_reuse=True,pre=True)(makeList)
    _str2bool_other_requirements=validator('other_requirements',allow_reuse=True,pre=True)(makeList)
    
    @property
    def date_of_offer(self):
        return self.offer_date.strftime("%b %d, %Y")
    
    @property
    def start_date_say(self):
        return self.work_start_date.strftime("%b %d, %Y")
    
    @property
    def requirements(self):
        return [r for r in [self.specific_edu_requirement,self.skill_experience_requirement,*self.other_requirements] if r is not None]

class PersonalAssess(BaseModel):
    work_experience_brief:str
    education_brief:str
    competency_brief:str
    language_brief:Optional[str]
    performance_remark:Optional[str]
    
    @property
    def why_qualified(self):
        qualifications=[self.work_experience_brief,self.education_brief,self.competency_brief,self.language_brief,self.performance_remark]
        return [q for q in qualifications if q is not None]



from datetime import date
from pydantic import BaseModel,validator
from dateutil.parser import parse,ParserError
from typing import List,Optional


class ID(BaseModel):
    variable_type:Optional[str]
    display_type:Optional[str]
    number:Optional[str]
    country:Optional[str]
    issue_date:Optional[date]
    expiry_date:Optional[date]

    @property
    def is_expired(self):
        return True if self.expiry_date and date.today()>=self.expiry_date else False
    
    def will_be_expired(self,the_day=date.today()):
        try:
            the_day=the_day if isinstance(the_day,date) else parse(the_day)
        except ParserError as err:
            raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        return True if self.expiry_date and the_day>=self.expiry_date else False
    
    @validator("expiry_date")
    def endDateBigger(cls,expiry_date,values):
        start_date=values.get('issue_date')
        if not start_date: return expiry_date
        the_date=expiry_date or date.today()
        if (the_date-start_date).days<=0:
            raise ValueError(f'Expiry date {the_date} is earlier than issue date {start_date} for the ID')
        expiry_date=expiry_date if expiry_date else "Present"
        return expiry_date
    

class IDs():
    def __init__(self,ids:List[ID]) -> None:
        self.ids=ids
    
    @property
    def passport(self):
        passport=[ id for id in self.ids if id.variable_type.lower()=='passport']
        return passport and passport[0]
    
    @property
    def national_id(self):
        national_id=[ id for id in self.ids if id.variable_type.lower()=='id']
        return national_id and national_id[0]
    
    @property
    def pr(self):
        pr=[ id for id in self.ids if id.variable_type.lower()=='pr']
        return pr and pr[0]
    
    


from pydantic import BaseModel
from termcolor import colored
from functools import reduce
from typing import List,Optional
from model.common.address import Address
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from datetime import date

class Family(BaseModel):
    last_name:str
    first_name:str	
    native_last_name:Optional[str]	
    native_first_name:Optional[str] 
    date_of_birth:date  
    date_of_death:Optional[date]
    place_of_birth:str
    birth_country:str 
    relationship:str
    
class Personal(BaseModel):
    last_name:str
    first_name:str
    native_last_name:str
    native_first_name:str
    date_of_birth:date	

class PRBackground(BaseModel):
    q1:bool
    q2:bool
    q3:bool
    q4:bool
    q5:bool
    q6:bool
    q7:bool
    q8:bool
    q9:bool
    q10:bool
    q11:bool
    details:Optional[str]

class Education(BaseModel):
    start_date:Optional[date]
    end_date:Optional[date]
    school_name:Optional[str]
    city:Optional[str]
    country:Optional[str]
    education_level:Optional[str]	
    field_of_study:Optional[str]

class History(BaseModel):
    start_date:date
    end_date:date	
    activity:str	
    city_and_country:str	
    status:str	
    name_of_company_or_school:str

class Member(BaseModel):
    start_date:date	
    end_date:date	
    organization_name:str	
    organization_type:str	
    position:str	
    city:str	
    country:str

class Government(BaseModel):
    start_date:date	
    end_date:date	
    country:str	
    department:str	
    position:str

class Military(BaseModel):
    start_date:date	
    end_date:date	
    location:str	
    province:str	
    country:str	
    service_detail:str	
    rank:str	
    combat_detail:Optional[str]	
    reason_for_end:Optional[str]

class AddressHistory(BaseModel):
    start_date:date	
    end_date:date	
    street_and_number:str	
    city:str	
    province:str	
    country:str	
    post_code:str    
        
class M5669Model(CommonModel):
    personal:Personal
    family:List[Family]
    prbackground:PRBackground
    education:List[Education]
    history:List[History]
    member:List[Member]
    government:List[Government]
    military:List[Military]
    addresshistory:List[AddressHistory]

    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            import os
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/pr.xlsx',
                path+'/template/excel/pa.xlsx'
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        super().__init__(excels,output_excel_file,globals())


from typing import List,Optional,Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.id import ID
from datetime import date
from pydantic import BaseModel, EmailStr,root_validator,validator
from model.common.utils import normalize

class PersonId(ID):
    pass

class PRCase(BaseModel):
    imm_program:str
    imm_category:str
    imm_under:Optional[str]
    communication_language:str
    interview_language:str
    need_translator:bool
    intended_province:str
    intended_city:str

class COR(BaseModel):
    start_date:date
    end_date:Union[date,None]
    country:str
    status:str
    
    def is_current(self):
        return not self.end_date

class Marriage(BaseModel):
    marital_status:str
    married_date:Optional[date]
    sp_last_name:Optional[str]
    sp_first_name:Optional[str]
    previous_married:bool
    pre_sp_last_name:Optional[str]
    pre_sp_first_name:Optional[str]
    pre_sp_dob:Optional[date]
    pre_relationship_type:Optional[str]
    pre_start_date:Optional[date]
    pre_end_date:Optional[date]

class Personal(BaseModel):
    last_name:str
    first_name:str
    used_last_name:Optional[str]
    used_first_name:Optional[str]
    sex:str
    height:int
    eye_color:str
    dob:date
    country_of_birth:str
    place_of_birth:str
    uci:Optional[str]
    citizen:str
    citizen2:Optional[str]
    native_language:str
    english_french:str
    which_one_better:Optional[str]
    language_test:bool
    current_occupation:str
    intended_occupation:Optional[str]
    email:EmailStr
    
    _normalize_used_first_name=validator('used_first_name',allow_reuse=True,check_fields=False)(normalize)
    _normalize_used_last_name=validator('used_last_name',allow_reuse=True,check_fields=False)(normalize)
    
    @root_validator
    def checkAnswers(cls,values):
        questions=['english_french']
        explanations=['which_one_better']
        qas=dict(zip(questions,explanations))
        for k,v in qas.items():  
            if values.get(k) and values.get(k)=="Both" and not values.get(v):
                    raise ValueError(f"Since {k} is true, but you did not answer the question {v} in info-personal sheet")
        return values

class Education(EducationBase):
    pass

class M0008Model(CommonModel):
    personal:Personal
    cor:List[COR]
    marriage:Marriage
    address:List[Address]
    phone:List[Phone]
    personid:List[PersonId]
    education:List[Education]
    prcase:PRCase
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            import os
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/pr.xlsx',
                path+'/template/excel/pa.xlsx'
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        super().__init__(excels,output_excel_file,globals())


# before '_' is sheet name
# variable as key, xpath of xml as value. 
data_model={
    'applicant':{
        "name":"page1/SectionA/Applicant/AppName",
        "marital_status":"page1/SectionA/Applicant/ChildMStatus",	
        "date_of_birth":"page1/SectionA/Applicant/AppDOB",	
        "birth_country":"page1/SectionA/Applicant/AppCOB",	
        "address":"page1/SectionA/Applicant/AppAddress",
        "occupation":"page1/SectionA/Applicant/AppOccupation"
    },
    'spouse':{
        "name":"page1/SectionA/Spouse/SpouseName",
        "marital_status":"page1/SectionA/Spouse/ChildMStatus",	
        "date_of_birth":"page1/SectionA/Spouse/SpouseDOB",	
        "birth_country":"page1/SectionA/Spouse/SpouseCOB",	
        "address":"page1/SectionA/Spouse/SpouseAddress",
        "occupation":"page1/SectionA/Spouse/SpouseOccupation",	
        "accompany_to_canada_yes":"page1/SectionA/Spouse/SpouseYes",
        "accompany_to_canada_no":"page1/SectionA/Spouse/SpouseNo"
    },
    'mother':{
        "name":"page1/SectionA/Mother/MotherName",
        "marital_status":"page1/SectionA/Mother/ChildMStatus",	
        "date_of_birth":"page1/SectionA/Mother/MotherDOB",	
        "birth_country":"page1/SectionA/Mother/MotherCOB",	
        "address":"page1/SectionA/Mother/MotherAddress",
        "occupation":"page1/SectionA/Mother/MotherOccupation",	
        "accompany_to_canada_yes":"page1/SectionA/Mother/MotherYes",
        "accompany_to_canada_no":"page1/SectionA/Mother/MotherNo"
    },
    'father':{
        "name":"page1/SectionA/Father/FatherName",
        "marital_status":"page1/SectionA/Father/ChildMStatus",	
        "date_of_birth":"page1/SectionA/Father/FatherDOB",	
        "birth_country":"page1/SectionA/Father/FatherCOB",	
        "address":"page1/SectionA/Father/FatherAddress",
        "occupation":"page1/SectionA/Father/FatherOccupation",	
        "accompany_to_canada_yes":"page1/SectionA/Father/FatherYes",
        "accompany_to_canada_no":"page1/SectionA/Father/FatherNo"
    },
    'children':{
        "name":"ChildName",
        "marital_status":"ChildMStatus",	
        "relationship":"ChildRelationship",
        "date_of_birth":"ChildDOB",	
        "birth_country":"ChildCOB",	
        "address":"ChildAddress",
        "occupation":"ChildOccupation",	
        "accompany_to_canada_yes":"ChildYes",
        "accompany_to_canada_no":"ChildNo"
    },
    'siblings':{
        "name":"ChildName",
        "marital_status":"ChildMStatus",	
        "relationship":"ChildRelationship",
        "date_of_birth":"ChildDOB",	
        "birth_country":"ChildCOB",	
        "address":"ChildAddress",
        "occupation":"ChildOccupation",	
        "accompany_to_canada_yes":"ChildYes",
        "accompany_to_canada_no":"ChildNo"
    }
}

convert_model={
    'tr_marital_status_pairs':{
            'applicant':['marital_status'],
            'spouse':['marital_status'],
            'mother':['marital_status'],
            'father':['marital_status'],
            'children':['marital_status'],
            'siblings':['marital_status'],
        }
    
}

remove_model={
            'spouse':['accompany_to_canada_yes','accompany_to_canada_no'],
            'mother':['accompany_to_canada_yes','accompany_to_canada_no'],
            'father':['accompany_to_canada_yes','accompany_to_canada_no']
        }


from pydantic import BaseModel,validator
from typing import Optional
from datetime import date
from model.common.utils import makeList
from pydantic.class_validators import root_validator
    
class JobofferBase(BaseModel):
    job_title:str
    noc:str
    hours:float
    wage_unit:str
    wage_rate:float
    ot_ratio:float
    permanent:bool
    job_duration:Optional[float]
    job_duration_unit:Optional[str]
    
    @root_validator
    def checkDuration(cls,values):
        if not values.get('permanent') and (not values.get('job_duration') or not values.get('job_duration_unit')):
                raise ValueError("Since it is not permanent job offer, so you have to specify the job duration and job duration unit in info-joboffer sheet")
        return values
    
    @property
    def term(self):
        if self.permanent:
            return "Permanent"
        else:
            if not self.job_duration or not self.job_duration_unit:
                raise ValueError('Since you claimed the job offer is not permanent, you must spcify the job duration and duration unit')
            return str(self.job_duration)+" "+self.job_duration_unit
    
    @property
    def full_part_time(self):
        return "full-time" if self.hours>=30 else 'part-time'
    
    @property
    def salary(self):
        return '{:,.1f}'.format(self.wage_rate) 
    
    @property
    def weekly_hours(self):
        return '{:,.1f}'.format(self.hours)
    
    @property    
    def hourly_rate(self):
        rate=0
        match self.wage_unit:
            case 'annually':
                rate=self.wage_rate/52/self.hours
            case 'monthly':
                rate= self.wage_rate*12/52/self.hours
            case 'weekly':
                rate= self.wage_rate/self.hours
            case 'hourly':
                rate= self.wage_rate
        return '{0:.4g}'.format(rate)

    @property    
    def hourly_rate_say(self):
        return f'${self.hourly_rate} per hour'
    
    @property    
    def overtime_rate(self):
        return float(self.hourly_rate)*self.ot_ratio
    
    @property
    def overtime_rate_say(self):
        return '${rate:,.0f} per hour'.format(rate=float(self.overtime_rate))
    
    @property
    def weekly_rate(self):
        rate=0
        match self.wage_unit:
            case 'annually':
                rate=self.wage_rate/52
            case 'monthly':
                rate= self.wage_rate*12/52
            case 'weekly':
                rate= self.wage_rate
            case 'hourly':
                rate= self.wage_rate*self.hours
        return '{0:.0f}'.format(rate)
    
    @property
    def weekly_rate_say(self):
        return '${rate:,.0f} per week'.format(rate=float(self.weekly_rate))
    
    @property
    def monthly_rate(self):
        rate=0
        match self.wage_unit:
            case 'annually':
                rate=self.wage_rate/12
            case 'monthly':
                rate= self.wage_rate
            case 'weekly':
                rate= self.wage_rate*52/12
            case 'hourly':
                rate= self.wage_rate*self.hours*52/12
        return '{0:.0f}'.format(rate)
    
    @property    
    def monthly_rate_say(self):
        return '${rate:,.0f} per month'.format(rate=float(self.monthly_rate))
    
    @property
    def annual_rate(self):
        rate=0
        match self.wage_unit:
            case 'annually':
                rate=self.wage_rate
            case 'monthly':
                rate= self.wage_rate*12
            case 'weekly':
                rate= self.wage_rate*52
            case 'hourly':
                rate= self.wage_rate*self.hours*52
        return '{0:.0f}'.format(rate)
    
    @property    
    def annual_rate_say(self):
        return '${rate:,.0f} per year'.format(rate=float(self.annual_rate))
    
    @property
    def requirements(self):
        reqs=[self.specific_edu_requirement,self.skill_experience_requirement,*self.other_requirements]
        return [req for req in reqs if req]




from pydantic import BaseModel,validator,EmailStr
from datetime import date
from typing import Optional,List
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.address import Address,Addresses
from model.common.utils import makeList
from pydantic.class_validators import root_validator

class ErAddress(Address):
    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()
    
class ErAddresses(Addresses):
    def __init__(self, address_list: List[Address]) -> None:
        super().__init__(address_list)
    

class General(EmployerBase):
    company_intro:str
    business_intro:str
    recruit_email:EmailStr
        
class JobOffer(JobofferBase):
    offer_date:date
    supervisor_name:Optional[str]
    supervisor_title:Optional[str]
    vacation_pay_weeks:int
    vacation_pay_percentage:float
    employer_rep:str
    employer_rep_title:str
    payment_way:str
    has_probation:bool
    probation_duration:Optional[int]
    duties_brief:str
    duties:list

    _str2bool_duties=validator('duties',allow_reuse=True,pre=True)(makeList)

    
    
    @root_validator
    def checkProbation(cls,values):
        if values.get('has_probation') and not values.get('probation_duration'):
                raise ValueError("Since it is has probation period, but you did not specify the probation duration in info-joboffer sheet")
        return values
    
    @property
    def date_of_offer(self):
        return self.offer_date.strftime("%b %d, %Y")
    
    @property
    def vacation_pay_percent(self):
        return '{:,.1f}%'.format(self.vacation_pay_percentage*100)
    


from typing import List
from model.common.commonmodel import CommonModel
from model.bcpnp.data import Bcpnp, General,JobOffer,ErAddress
from model.common.jobposition import Position
from model.common.rcic import Rcic
from model.common.advertisement import Advertisement,Advertisements,InterviewRecord,InterviewRecords,RecruitmentSummary
from model.common.person import Person,PersonalAssess
from model.common.address import Addresses
from model.common.wordmaker import WordMaker
import os

class Personal(Person):
    def __str__(self):
        return self.full_name
        
class EmployerTrainingModel(CommonModel):
    eraddress:List[ErAddress]
    general:General
    position:Position
    personal:Personal
    joboffer:JobOffer
    personalassess:PersonalAssess
    bcpnp:Bcpnp
    rcic:Rcic
    advertisement:List[Advertisement]
    interviewrecord:List[InterviewRecord]
    recruitmentsummary:RecruitmentSummary

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            import os
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/er.xlsx',
                path+"/template/excel/pa.xlsx",
                path+"/template/excel/recruitment.xlsx",
                path+"/template/excel/bcpnp.xlsx",
                path+"/template/excel/rep.xlsx"
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())
    
    @property
    def work_location(self):
        addresses=Addresses(self.eraddress)
        return addresses.working
    
    @property
    def selected_contact(self):
        contacts=Contacts(self.contact)
        return contacts.preferredContact
    
    @property
    def summary(self):
        return InterviewRecords(self.interviewrecord)
    
    @property
    def advertisements(self):
        return Advertisements(self.advertisement)
    
    @property
    def person(self):
        return {
            "full_name":self.personal.full_name,
            "attributive":self.personal.attributive,
            "object":self.personal.object,
            "subject":self.personal.subject,
            'short_name':self.personal.short_name
        }

class EmployerTrainingDocxAdaptor():
    def __init__(self,employer_training_obj:EmployerTrainingModel):
        self.employer_training_obj=employer_training_obj
        
    def re_generate_dict(self):
        summary_info={
            'resume_num':self.employer_training_obj.summary.resume_num,
            'canadian_num':self.employer_training_obj.summary.canadian_num,
            'unknown_num':self.employer_training_obj.summary.unknown_num,
            "foreigner_num":self.employer_training_obj.summary.foreigner_num,
            "total_canadian":self.employer_training_obj.summary.total_canadian,
            "total_interviewed_canadians":self.employer_training_obj.summary.total_interviewed_canadians,
            "canadian_records":self.employer_training_obj.summary.canadian_records,
            "advertisement":self.employer_training_obj.advertisements,
            'personal':self.employer_training_obj.person,
            "date_of_offer":self.employer_training_obj.joboffer.date_of_offer,
            "work_start_date":self.employer_training_obj.joboffer.start_date_say,
            'joboffer_date':self.employer_training_obj.joboffer.date_of_offer,
            'work_location':self.employer_training_obj.work_location
        }
        return {**self.employer_training_obj.dict(),**summary_info}

    def make(self,output_docx):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/word/bcpnp_facts_brief_employer.docx"            
        wm=WordMaker(template_path,self.re_generate_dict(),output_docx)
        wm.make()



from dataclasses import dataclass
from typing import List
from model.common.commonmodel import CommonModel
from model.lmia.data import LmiaCase,Finance,Lmi,Emp5626,Emp5627,Emp5593,General,JobOffer,PersonalAssess,Rcic
from model.common.jobposition import Position
from model.common.advertisement import InterviewRecord,Advertisement
from model.common.person import Person
from model.common.wordmaker import WordMaker
import os

class Personal(Person):
    def __str__(self):
        return self.full_name
        
class EmployerTrainingModel(CommonModel):
    general:General
    position:Position
    personal:Personal
    joboffer:JobOffer
    personalassess:PersonalAssess
    lmiacase:LmiaCase
    finance:List[Finance]
    lmi:Lmi
    rcic:Rcic
    advertisement:List[Advertisement]
    interviewrecord:List[InterviewRecord]
    
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=[
                path+'/template/excel/er.xlsx',
                path+"/template/excel/pa.xlsx",
                path+"/template/excel/recruitment.xlsx",
                path+"/template/excel/lmia.xlsx",
                path+"/template/excel/rep.xlsx"
            ]
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())
    
    def makeDocx(self,output_docx):
        path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path=path+"/template/word/lmia_employer_training.docx"            
        wm=WordMaker(template_path,self,output_docx)
        wm.make()
    
    
class M5593EmployerTrainingModel(EmployerTrainingModel):
    pass
    
class M5626EmployerTrainingModel(EmployerTrainingModel):
    pass
    
class M5627EmployerTrainingModel(EmployerTrainingModel):
    pass




from setuptools import setup
from setuptools import find_packages,find_namespace_packages

# Load the README file.
with open(file="README.md", mode="r") as readme_handle:
    long_description = readme_handle.read()

setup(
    # Define the library name, this is what is used along with `pip install`.
    name='imm-case-model',

    # Define the author of the repository.
    author='Jacky Zhang',

    # Define the Author's email, so people know who to reach out to.
    author_email='jackyzhang1969@gmail.com',

    # Define the version of this library.
    # Read this as
    #   - MAJOR VERSION 0
    #   - MINOR VERSION 1
    #   - MAINTENANCE VERSION 0
    version='1.0.19',

    # Here is a small description of the library. This appears
    # when someone searches for the library on https://pypi.org/search.
    description='A python client library used to describe program model and get validated data.',

    # I have a long description but that will just be my README
    # file, note the variable up above where I read the file.
    long_description=long_description,

    # This will specify that the long description is MARKDOWN.
    long_description_content_type="text/markdown",

    # Here is the URL where you can find the code, in this case on GitHub.
    url='',

    # These are the dependencies the library needs in order to run.
    install_requires=[
       'imm-source'
    ],

    # Here are the keywords of my library.
    keywords='Canada, Immigration, Data, Model, Programs, APIs',
    
    # here are the packages I want "build."
    packages=find_packages(exclude=('test*','try*')),

    # here we specify any package data.


    # I also have some package data, like photos and JSON files, so
    # I want to include those as well.
    include_package_data=True,

    # Here I can specify the python version necessary to run this library.
    python_requires='>=3.7',

    # Additional classifiers that give some characteristics about the package.
    # For a complete list go to https://pypi.org/classifiers/.
    classifiers=[

        # I can say what phase of development my library is in.
        'Development Status :: 3 - Alpha',

        # Here I'll add the audience this library is intended for.
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Financial and Insurance Industry',

        # Here I'll define the license that guides my library.
        'License :: OSI Approved :: MIT License',

        # Here I'll note that package was written in English.
        'Natural Language :: English',

        # Here I'll note that any operating system can use it.
        'Operating System :: OS Independent',

        # Here I'll specify the version of Python it uses.
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',

        # Here are the topics that my library covers.
        # 'Topic :: Data',
        # 'Topic :: Immigration',
        # 'Topic :: Canada'

    ],
    # Here defines the console scripts. Terminal command, and it's source. 
    # entry_points={
    #     'console_scripts':[
    #         'ds=source.datasource:main'  # ds是命令行命令， source.source:main是对应的函数所在包中的位置
    #     ]
    # }
)


"""
This class is for code and name convert. Usually used for country, province,city, etc... to convert between name and code
"""

class NameCodeConvert(object):
    # initialize with a dict including name and code
    def __init__(self,d:dict):
        self.d=d

    def getName(self,code):
        c={v:k for k,v in self.d.items()}
        return c.get(str(code))
    
    def getCode(self,name):
        # remove additional space and trim left and right
        name=' '.join(name.split()).strip()
        for item_name, code in self.d.items():
            if name in item_name.lower():
                return code
#TODO: 此处可以考虑增加模糊选项，如果有错，提供对应的输入建议。 参考noc的输入检查




from model.tr.d0104 import data_model
from model.tr.pdfformmodel import PdfFormModel


class F0104Model(PdfFormModel):
    
    def __init__(self,xml_file):
        # pass the globals variables to base class
        model={'data_model':data_model,'convert_model':{},'remove_model':{}}
        super().__init__(xml_file,model)

    def _getData(self):
        # work
        data_pairs={
            'employment':"Page1/EmploymentSub/EmploymentTbl",
            "education":"Page1/EducationSub/EducationTbl",
            "travel":"Page1/TravelSub/TravelTbl"
        }
        
        for key,parent_xpath in data_pairs.items():
            the_parent_xpath = [e for e in self.tree.iterfind(parent_xpath)][0]
            rows = []
            for i in range(1,100):
                employment_item =the_parent_xpath.find('Row'+str(i))
                if not employment_item:
                    break
                employment = self._getSubItem(employment_item, data_model[key])
                dates=self._getSubItem(employment_item,{'dates':"TimePeriodFrom"})
                employment={**dates,**employment}
                rows.append(employment)
            self.data[key] = rows

        # travel re-ordering
        travels=self.data['travel']
        temp=[]
        for travel in travels:
            travel={k:travel.get(k) for k in ['start_date','end_date','length','destination','purpose']}
            temp.append(travel)
        self.data['travel']=temp
        
    def _getSubItem(self,element,variable_xpath_pairs):
            temp_dict={}
            for variable,xpath in variable_xpath_pairs.items():
                try:
                    the_element = element.findall(xpath)
                    if len(the_element)>1:
                        start_date = the_element[0].text
                        temp_dict['start_date']=self._formatDate(start_date)
                        end_date = the_element[1].text
                        temp_dict['end_date']=self._formatDate(end_date)
                    elif len(the_element)==1:
                        value = the_element[0].text
                        temp_dict[variable]=value
                    else:
                        print(f"{xpath} is invalid xpath")    
                except:
                    print(f"{xpath} is invalid xpath")
            return temp_dict

    def _formatDate(self,input_date):
        y,m=input_date.split('-')[1],input_date.split('-')[0]
        return y+'-'+m+"-"+"01"
    
    def _convert(self):
        pass
    
    def _assembl(self):
        pass
    
    def _remove(self):
        pass
    


from abc import abstractclassmethod
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element,SubElement

"""     
1. 根据表格字典，构建所有节点
2. 根据excel数据，以及program定义的结构，将原始数据按照对应的数据字典的keys来拆分
3. switch数据字典，根据xpath，找到数据的key value，然后给xpath的元素赋值

"""
class XmlMaker(object):
    def __init__(self,model,root_tag:str,source_data):
        self.model=model
        self.source_data=source_data
        # 根据excel来的数据，改造model,成为直接匹配xml的格式。 
        # self.model=self.rebuild_model()
        # 构建树根
        self.root=ET.Element(root_tag)
        # 此处的paths是在已经重构的model下，所有的xpath都是绝对路径
        self.paths=self.getPaths()
        # 够条件xml树
        self.makeTree()
        #根据重构的model和source_data来赋值
        self.fill()
        
    # 不同的项目重写该方法。
    @abstractclassmethod
    def rebuild_model(self):
        raise NotImplementedError('rebuild_model method in class XmlMaker  not implemented...')
        
    def makeTree(self):
        for path_pairs in self.paths.values():
            for path in path_pairs.values():
                self.make_tree(self.root,path)
        
    def getPaths(self):
        path_model={}
        for key,value_pairs in self.model.items():
            path_model[key]={}
            for variable,path in value_pairs.items():
                path=path.split('/')
                path.reverse()
                path_model[key][variable]=path
        return path_model

    # 利用递归来构建无穷可能的树
    def make_tree(self,root,data:list):
        d=data.pop()
        parent=root
        new_element=parent.find(d)
        if isinstance(new_element,Element):
            parent=new_element
        else:
            parent=SubElement(parent,d)
        if len(data)>0:
            self.make_tree(parent,data)
        else:
            return 

    # fill xml tag text with value
    def fill(self):
        for sheet, xpath_pairs in self.model.items():
            for variable, xpath in xpath_pairs.items():
                s=self.root.find(xpath)
                s.text=str(self.source_data[sheet][variable])
    
    def write(self,output_xml_filename):
        tree = ET.ElementTree(self.root)
        tree.write(output_xml_filename)





from model.tr.d5708 import data_model, convert_model,remove_model
from model.tr.pdfformmodel import PdfFormModel

class F5708Model(PdfFormModel):

    def __init__(self,xml_file):
            # pass the globals variables to base class
        model={'data_model':data_model,'convert_model':convert_model,'remove_model':remove_model}
        super().__init__(xml_file,model)        
    
    def _getSpecial(self):
        # work 
        occupations=[]
        for i in range(1,100):
            occupation_item=self.tree.find('Page3/Employment/EmpRec'+str(i))
            if occupation_item is None:
                break
            occupation=self._getSubItem(occupation_item,data_model['occupation'])
            occupations.append(occupation)

        self.data['occupations']=occupations

    



data_model={
    'trcasein':{
        'service_in':'Page1/PersonalDetails/ServiceIn/ServiceIn',
        "application_purpose1":"Page1/PersonalDetails/ApplyingFor/RestoreStat",
        "application_purpose2":"Page1/PersonalDetails/ApplyingFor/Extend",
        "application_purpose3":"Page1/PersonalDetails/ApplyingFor/TRP",
        "original_entry_date":"Page3/ComingIntoCda/OrigEntry/DateLastEntry",
        "original_entry_place":"Page3/ComingIntoCda/OrigEntry/Place",
        "original_purpose":"Page3/ComingIntoCda/PurposeOfVisit/PurposeOfVisit",
        "original_other_reason":"Page3/ComingIntoCda/PurposeOfVisit/Other",
        "most_recent_entry_date":"Page3/ComingIntoCda/RecentEntry/DateLastEntry",
        "most_recent_entry_place":"Page3/ComingIntoCda/RecentEntry/Place",
        "doc_number":"Page3/ComingIntoCda/PrevDocNum/docNum",
        "is_spouse_canadian":"Page1/MaritalStatus/d/SpouseStatus",
        "consent_of_info_release":"Page4/Signature/FutureComm",
        "submission_date":"Page4/Signature/C1CertificateIssueDate"
    },
    "vrincanada":{
        "visit_purpose":"Page3/DetailsOfVisit/Purpose/Purpose",
        "duration_from":"Page3/DetailsOfVisit/Purpose/Stay/FromDate",
        "duration_to":"Page3/DetailsOfVisit/Purpose/Stay/ToDate",
        "funds_available":"Page3/DetailsOfVisit/Funds/FundsAvail",
        "name1":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec1/Name",
        "relationship1":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec1/Relationship",
        "address1":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec1/Addr",
        "name2":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec2/Name",
        "relationship":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec2/Relationship",
        "address2":"Page3/DetailsOfVisit/WillVisit/VisitList/Rec2/Addr"
    },
    'personal':{
        'uci':"Page1/PersonalDetails/ServiceIn/UCIClientID",
        'last_name':"Page1/PersonalDetails/Name/FamilyName",
        'first_name':"Page1/PersonalDetails/Name/GivenName",
        'has_alias_name':"Page1/PersonalDetails/AliasName/AliasNameIndicator/AliasNameIndicator",
        'used_last_name':"Page1/PersonalDetails/AliasName/AliasFamilyName",
        'used_first_name':"Page1/PersonalDetails/AliasName/AliasGivenName",
        'sex':"Page1/PersonalDetails/q3-4-5/sex/Sex",
        'dob_year':"Page1/PersonalDetails/q3-4-5/dob/DOBYear",
        'dob_month':"Page1/PersonalDetails/q3-4-5/dob/DOBMonth",
        'dob_day':"Page1/PersonalDetails/q3-4-5/dob/DOBDay",
        'place_of_birth':"Page1/PersonalDetails/q3-4-5/pob/PlaceBirthCity",
        'country_of_birth':"Page1/PersonalDetails/q3-4-5/pob/PlaceBirthCountry",
        'citizen':"Page1/PersonalDetails/Citizenship/Citizenship",
        "native_language":"Page2/Languages/nativeLang",
        "english_french":"Page2/Languages/communicateLang",
        "which_one_better":"Page2/Languages/FreqLang",
        "language_test":"Page2/Languages/LangTestIndicator",
        "email":"Page2/ContactInformation/q5-6/Email"
    },
    #country of residence: current and past for over 6 months. table model
    'cor':{
        'current_cor_country':"Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Country",
        'current_cor_status':"Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Status",
        'current_cor_other':"Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Other",
        'current_cor_start_date':"Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/FromDate",
        'current_cor_end_date':"Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/ToDate",
        "previous_cor_country1":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Country",
        "previous_cor_status1":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Status",
        "previous_cor_other1":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Other",
        "previous_cor_start_date1":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/FromDate",
        "previous_cor_end_date1":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/ToDate",
        "previous_cor_country2":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Country",
        "previous_cor_status2":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Status",
        "previous_cor_other2":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Other",
        "previous_cor_start_date2":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/FromDate",
        "previous_cor_end_date2":"Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/ToDate"
    },

    # marriage
    'marriage':{
        "marital_status":"Page1/MaritalStatus/Current/MaritalStatus",
        "married_date":"Page1/MaritalStatus/Current/b/DateOfMarriage",
        "sp_last_name":"Page1/MaritalStatus/Current/c/FamilyName",
        "sp_first_name":"Page1/MaritalStatus/Current/c/GivenName",
        "previous_married":"Page2/MaritalStatus/PrevMarriage/PrevMarriedIndicator",
        "pre_sp_last_name":"Page2/MaritalStatus/PrevMarriage/PMFamilyName",
        "pre_sp_first_name":"Page2/MaritalStatus/PrevMarriage/PMGivenName",
        "pre_relationship_type":"Page2/MaritalStatus/PrevMarriage/TypeOfRelationship",
        "pre_sp_dob_year":"Page2/MaritalStatus/PrevMarriage/dob/DOBYear",
        "pre_sp_dob_month":"Page2/MaritalStatus/PrevMarriage/dob/DOBMonth",
        "pre_sp_dob_day":"Page2/MaritalStatus/PrevMarriage/dob/DOBDay",
        "pre_start_date":"Page2/MaritalStatus/PrevMarriage/From/FromDate",
        "pre_end_date":"Page2/MaritalStatus/PrevMarriage/To/ToDate"
    },

    'passport':{
        "number":"Page2/Passport/PassportNum",
        "country":"Page2/Passport/CountryofIssue",
        "issue_date":"Page2/Passport/IssueDate",
        "expiry_date":"Page2/Passport/ExpiryDate"
    },
    'national_id':{
        "has_id":"Page2/natID/q1/natIDIndicator",
        "number":"Page2/natID/natIDdocs/DocNum/DocNum",
        "country":"Page2/natID/natIDdocs/CountryofIssue/CountryofIssue",
        "issue_date":"Page2/natID/natIDdocs/IssueDate/IssueDate",
        "expiry_date":"Page2/natID/natIDdocs/ExpiryDate"
    },
    'us_pr':{
        "has_id":"Page2/USCard/q1/usCardIndicator",
        "number":"Page2/USCard/usCarddocs/DocNum/DocNum",
        "expiry_date":"Page2/USCard/usCarddocs/ExpiryDate"
    },

    'mailing_address':{
        'po_box':"Page2/ContactInformation/Mailing/AddrLine1/POBox",
        'unit':"Page2/ContactInformation/Mailing/AddrLine1/AptUnit",
        'street_number':"Page2/ContactInformation/Mailing/AddrLine1/StreetNum",
        'street_name':"Page2/ContactInformation/Mailing/AddrLine1/Streetname",
        'city':"Page2/ContactInformation/Mailing/AddrLine2/City",
        'country':"Page2/ContactInformation/Mailing/AddrLine2/Country",
        'province':"Page2/ContactInformation/Mailing/AddrLine2/Prov",
        'post_code':"Page2/ContactInformation/Mailing/AddrLine2/PostalCode",
        'district':"Page2/ContactInformation/Mailing/AddrLine2/District"
    },

    'residential_address':{
        'same_as_mailing':"Page2/ContactInformation/Resi/SameAsAddr/SameAsMailingInd",
        'unit':"Page2/ContactInformation/Resi/AddrLine1/AptUnit",
        'street_number':"Page2/ContactInformation/Resi/AddrLine1/StreetNum",
        'street_name':"Page2/ContactInformation/Resi/AddrLine1/Streetname",
        'city':"Page2/ContactInformation/Resi/AddrLine2/City",
        'country':"Page2/ContactInformation/Resi/AddrLine2/Country",
        'province':"Page2/ContactInformation/Resi/AddrLine2/Prov",
        'post_code':"Page2/ContactInformation/Resi/AddrLine2/PostalCode",
        'district':"Page2/ContactInformation/Resi/AddrLine2/District"
    },

    'phone':{
        "variable_type":"Page2/ContactInformation/q3-4/Phone/Type",
        "canada_us":"Page2/ContactInformation/q3-4/Phone/CanOtherInd/CanadaUS",
        "other":"Page2/ContactInformation/q3-4/Phone/CanOtherInd/Other",
        'ext':"Page2/ContactInformation/q3-4/Phone/NumberExt",
        'country_code':"Page2/ContactInformation/q3-4/Phone/NumberCountry",
        'number':"Page2/ContactInformation/q3-4/Phone/ActualNumber",
        'na_area':"Page2/ContactInformation/q3-4/Phone/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/q3-4/Phone/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/q3-4/Phone/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/q3-4/Phone/IntlNumber/IntlNumber"
    },

    'altphone':{
        "variable_type":"Page2/ContactInformation/q3-4/AltPhone/Type",
        "canada_us":"Page2/ContactInformation/q3-4/AltPhone/CanOtherInd/CanadaUS",
        "other":"Page2/ContactInformation/q3-4/AltPhone/CanOtherInd/Other",
        'ext':"Page2/ContactInformation/q3-4/AltPhone/NumberExt",
        'country_code':"Page2/ContactInformation/q3-4/AltPhone/NumberCountry",
        'number':"Page2/ContactInformation/q3-4/AltPhone/ActualNumber",
        'na_area':"Page2/ContactInformation/q3-4/AltPhone/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/q3-4/AltPhone/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/q3-4/AltPhone/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/q3-4/AltPhone/IntlNumber/IntlNumber"
    },

    'fax':{
        "canada_us":"Page2/ContactInformation/q5-6/Fax/CanOtherInd/CanadaUS",
        "other":"Page2/ContactInformation/q5-6/Fax/CanOtherInd/Other",
        'ext':"Page2/ContactInformation/q5-6/Fax/NumberExt",
        'country_code':"Page2/ContactInformation/q5-6/Fax/NumberCountry",
        'number':"Page2/ContactInformation/q5-6/Fax/ActualNumber",
        'na_area':"Page2/ContactInformation/q5-6/Fax/NANumber/AreaCode",
        'na_first_three':"Page2/ContactInformation/q5-6/Fax/NANumber/FirstThree",
        'na_last_five':"Page2/ContactInformation/q5-6/Fax/NANumber/LastFive",
        'int_number':"Page2/ContactInformation/q5-6/Fax/IntlNumber/IntlNumber"
    },

    'education':{
        "start_date_year":"Page3/Education/EduLine1/From/YYYY",
        "start_date_month":"Page3/Education/EduLine1/From/MM",
        "end_date_year":"Page3/Education/EduLine2/To/YYYY",
        "end_date_month":"Page3/Education/EduLine2/To/MM",
        "school_name":"Page3/Education/EduLine1/School",
        "field_of_study":"Page3/Education/EduLine1/FieldOfStudy",
        "city":"Page3/Education/EduLine2/City",
        "country":"Page3/Education/EduLine2/Country",
        "province":"Page3/Education/EduLine2/Prov"
    },
    # occupation is a special dict, and will be handled seperately
    'occupation':{
        "start_date_year":"Line1/From/YYYY",
        "start_date_month":"Line1/From/MM",
        "end_date_year":"Line2/To/YYYY",
        "end_date_month":"Line2/To/MM",
        "job_title":"Line1/Occupation",
        "company":"Line1/Employer",
        "city":"Line2/City",
        "province":"Line2/ProvState",
        "country":"Line2/Country"
    },

    'trbackground':{
        "q1a":"Page4/BackgroundInfo/HealthQ/qANY",
        "q1b":"Page4/BackgroundInfo/HealthQ/qBNY",
        "q1c":"Page4/BackgroundInfo/HealthQ/MedicalDetails",
        "q2a":"Page4/BackgroundInfo/PrevApplied/qANY",
        "q2b":"Page4/BackgroundInfo/PrevApplied/qBNY",
        "q2c":"Page4/BackgroundInfo/PrevApplied/qCNY",
        "q2d":"Page4/BackgroundInfo/PrevApplied/refusedDetails",
        "q3a":"Page4/BackgroundInfo/Criminal/qANY",
        "q3b":"Page4/BackgroundInfo/Criminal/refusedDetails",
        "q4a":"Page4/BackgroundInfo/Military/qANY",
        "q4b":"Page4/BackgroundInfo/Military/militaryServiceDetails",
        "q5":"Page4/BackgroundInfo/GovPosition/qGovtNY",
        "q6":"Page4/BackgroundInfo/Illtreatment/qWitnessNY"
    }
}


convert_model = {
    "country_pairs": {
        "trcase": ["applying_country"],
        "personal": ['country_of_birth', 'citizen'],
        'passport': ['country'],
        'national_id': ['country'],
        'education': ['country'],
        'cor': ['current_cor_country', 'previous_cor_country1', 'previous_cor_country2']
    },
    'canada_province_pairs': {
        'vrincanada': ["province"],
        'address': ['province'],
        'education': ['province'],
        'employment': ['province']
    },
    'language_pairs': {
        'personal': ['native_language']
    },
    'phone_type_pairs': {
        'phone': ["variable_type"],
        'altphone': ["variable_type"]
    },

    'tr_portal_study_level_pairs': {
        'sp': ["study_level"]
    },
    'tr_portal_study_field_pairs': {
        'sp': ["study_field"]
    },

    'tr_canada_status_pairs': {
        'cor': ['current_cor_status', 'previous_cor_status1', 'previous_cor_status2']
    },
    'tr_marital_status_pairs': {
        'marriage': ['marital_status', 'pre_relationship_type']
    }
}

remove_model = {
    'personal': ['dob_year', 'dob_month', 'dob_day'],
    'marriage': ['pre_sp_dob_year', 'pre_sp_dob_month', 'pre_sp_dob_day']
}



from pydantic import BaseModel,validator,EmailStr
from datetime import date
from model.common.utils import makeList,Duration
from model.common.person import Person
from model.common.id import ID,IDs
from typing import Optional,List
from model.common.employmentbase import EmploymentBase

class Employment(EmploymentBase):
    department:Optional[str]
    duties:list
    company_brief:str
    fullname_of_certificate_provider:Optional[str]
    position_of_certificate_provider:Optional[str]
    department_of_certificate_provider:Optional[str]
    phone_of_certificate_provider:Optional[str]
    email_of_certificate_provider:Optional[EmailStr]
    employment_certificate:bool
    
    _normalize_duties=validator('duties',allow_reuse=True,pre=True)(makeList)
    
